import json
import traceback

from dust import logger
from dust.httpservices import SVCINFO_MODULE, DustResultType
from dust.httpservices.servicebase import ServiceBase

MODULE_ACTIONS = "ACTIONS";

_services = []

class ActionController(ServiceBase):
    def __init__(self, *service_names):
        super().__init__(MODULE_ACTIONS)

        for service_name in service_names:
            _services.append(ServiceBase.get_service(service_name))

    def relay_action(self, params, request, response):
        ret = False

        name = params.get(SVCINFO_MODULE)

        services = None
        if name is None:
            services = _services
        else:
            services = []
            services.append(ServiceBase.get_service(name))

        for service in services:
            ret = ret or self.__relay_action(service, params, request, response)

        return ret

    def __relay_action(self, service, params, request, response):
        ret = False

        if service:
            try:
                result = service.do_process(params, request, response, immediate=True)
                if result.is_readon():
                    result = service.do_process(params, request, response, immediate=False)

                ret = result == DustResultType.ACCEPT
            except:
                traceback.print_exc()
                pass
            finally:
                pass

        return ret

    def do_process(self, params, request, response, immediate=True):
        ret = DustResultType.REJECT

        resp = {}

        for service in _services:
            if not service.do_process(params, request, resp, immediate).is_reject():
                service_name = service.get_modulename()
                ret = DustResultType.ACCEPT
                response[service_name] = resp
                resp = {}

        if ret != DustResultType.REJECT:
            self.log(response)

        return ret

    def log(self, action):
        logger().debug(json.dumps(action, indent=4))


import traceback
from jinja2 import Template

from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps
from dust.entity import Store, Entity

UNIT_TEMLATES = "templates"
UNIT_TEMLATES_META = "templates_meta"
UNIT_ID = 9
UNIT_META_ID = 10

class TemplateMeta(MetaProps):
    name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 100)
    text = (Datatypes.STRING, ValueTypes.MAP, 2, 101)

class TemplateTypes(FieldProps):
    template = (UNIT_TEMLATES_META, TemplateMeta, 1)

Store.create_unit(UNIT_TEMLATES, UNIT_ID)
Store.load_types_from_enum(TemplateTypes, UNIT_META_ID)

def create_template(name, text):
    template = Store.access(Operation.GET, None, UNIT_TEMLATES, None, TemplateTypes.template)
    template.access(Operation.SET, name, TemplateMeta.name)
    template.access(Operation.SET, text, TemplateMeta.text)
    return template

def render_template(template, **kwargs):
    try: 
        template = Template(template.access(Operation.GET, None, TemplateMeta.text))
        return template.render(**kwargs)
    except:
        traceback.print_exc()





import mysql.connector
from mysql.connector import errorcode
import io
import traceback
import os
import datetime

FORMAT_DB_DATETIME = "%Y-%m-%d %H:%M:%S"

DEFAULT_MAX_PACKET_SIZE = 4 * 1024 * 1024
DEFAUTL_MAX_NUM_OF_INSERTS = 1000
DEFAULT_SPLIT_DUMP_FILES = True

def _create_connection():
    conn = None
    try:
        conn = mysql.connector.connect(user=MYSQL_USER, password=MYSQL_PASSWORD, host=MYSQL_HOST, database=MYSQL_DB, charset='utf8', use_unicode=True)
    except mysql.connector.Error as err:
        if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
            print("Something is wrong with your user name or password")
        elif err.errno == errorcode.ER_BAD_DB_ERROR:
            print("Database does not exist")
        else:
            print(err)

    return conn

def _close_connection(conn):
    if conn:
        conn.commit()
        conn.close()

def _create_cursor(conn, prepared=False, buffered=True):
    if conn:
        return conn.cursor(buffered=buffered, prepared=prepared)

def _close_cursor(c):
    if c:
        c.close()

def _start_dump_file(stream, file_counter):
    stream.write(f"-- MySQL dump (pydust), file_counter: {file_counter}\n"+
                    "--\n"+
                    f"-- Host: {os.environ.get('HOSTNAME')}    Database: {MYSQL_DB}\n"+
                    "-- ------------------------------------------------------\n"+
                    f"-- Server version	{mysql_version}\n\n")

    stream.write("/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT */;\n"+
                    "/*!40101 SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS */;\n"+
                    "/*!40101 SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION */;\n"+
                    "/*!50503 SET NAMES utf8mb4 */;\n"+
                    "/*!40103 SET @OLD_TIME_ZONE=@@TIME_ZONE */;\n"+
                    "/*!40103 SET TIME_ZONE='+00:00' */;\n"+
                    "/*!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 */;\n"+
                    "/*!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 */;\n"+
                    "/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO' */;\n"+
                    "/*!40111 SET @OLD_SQL_NOTES=@@SQL_NOTES, SQL_NOTES=0 */;\n\n")

def _end_dump_file(stream):
    stream.write("/*!40101 SET SQL_MODE=@OLD_SQL_MODE */;\n"+
                    "/*!40014 SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS */;\n"+
                    "/*!40014 SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS */;\n"+
                    "/*!40101 SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT */;\n"+
                    "/*!40101 SET CHARACTER_SET_RESULTS=@OLD_CHARACTER_SET_RESULTS */;\n"+
                    "/*!40101 SET COLLATION_CONNECTION=@OLD_COLLATION_CONNECTION */;\n"+
                    "/*!40111 SET SQL_NOTES=@OLD_SQL_NOTES */;\n\n"+
                    f"-- Dump completed on {datetime.datetime.strftime(datetime.datetime.now(), FORMAT_DB_DATETIME)}\n")

conn = _create_connection()

try:
    file_counter = 0

    c = _create_cursor(conn, buffered=False)

    max_packet_size = int(os.environ.get("MYSQL_MAX_ALLOWED_PACKET_SIZE", DEFAULT_MAX_PACKET_SIZE))
    max_num_of_inserts = int(os.environ.get("MYSQL_MAX_NUM_OF_INSERTS", DEFAUTL_MAX_NUM_OF_INSERTS))
    split_files = bool(os.environ.get("MYSQL_SPLIT_DUMP_FILES", DEFAULT_SPLIT_DUMP_FILES))
    
    mysql_version = ""

    c.execute("SHOW VARIABLES LIKE 'version'")
    for vars in c.fetchall():
        mysql_version = vars[1]

    c.execute("SHOW TABLES")
    tables = []
    for table in c.fetchall():
        tables.append(table[0])

    # Lower the packet size slightly
    max_packet_size -= 1000

    for table in tables:
        stream = open("gcp_trivy_scanner_{}.sql".format(str(file_counter).zfill(6)), "w")
        _start_dump_file(stream, file_counter)
    
        stream.write("--\n"+
                    f"-- Table structure for table `{table}`\n"+
                    "--\n\n")
        stream.write("DROP TABLE IF EXISTS `" + str(table) + "`;\n")

        stream.write("/*!40101 SET @saved_cs_client     = @@character_set_client */;\n"+
                    "/*!50503 SET character_set_client = utf8mb4 */;")

        c.execute("SHOW CREATE TABLE `" + str(table) + "`;")
        stream.write("\n" + str(c.fetchone()[1]) + ";\n");

        stream.write("/*!40101 SET character_set_client = @saved_cs_client */;\n\n")

        stream.write("--\n"+
                    f"-- Dumping data for table `{table}`\n"+
                    "--\n\n"+
                    f"LOCK TABLES `{table}` WRITE;\n"+
                    f"/*!40000 ALTER TABLE `{table}` DISABLE KEYS */;\n")

        # Get fields:
        c.execute("DESCRIBE `" + str(table) + "`;")
        fields = []
        for field in c.fetchall():
            field_type = field[1]
            if field_type.lower() in ["binary", "varbinary", "blob", "mediumblob", "longblob"]:
                fields.append((f"HEX({field[0]})", field[0], True)) 
            else:
                fields.append((field[0], field[0], False)) 


        c.execute("SELECT {} FROM `{}`;".format(",".join(f[0] for f in fields), table))
        row = c.fetchone()

        row_counter = 0
        packet_length  = 0

        empty_table = True

        if row is not None:
            empty_table = False
            insert_str = "INSERT INTO `{}` ({}) VALUES ".format(table, ",".join(f[1] for f in fields))
            packet_length = len(insert_str)
            stream.write(insert_str)
            first_row = True

        while row is not None:
            row_stream = io.StringIO("")

            if not first_row:
                row_stream.write(",")

            row_stream.write("(")
            first = True
            for field_idx in range(len(fields)):
                field = fields[field_idx]
                if not first:
                    row_stream.write(",");
                if row[field_idx] is None:
                    row_stream.write("NULL")
                elif field[2]:
                    row_stream.write(f"UNHEX(\"{row[field_idx]}\")")
                elif isinstance(row[field_idx], str):
                    escaped_value = row[field_idx].replace('\\','\\\\').replace('"','\\"')
                    row_stream.write(f"\"{escaped_value}\"")
                else:
                    row_stream.write(f"\"{row[field_idx]}\"")
                first = False
            row_stream.write(")")

            row_str = row_stream.getvalue()
            if packet_length + len(row_str) > max_packet_size or ( max_num_of_inserts > 0 and row_counter > max_num_of_inserts ):
                row_counter = 0
                stream.write(";\n")

                if split_files:
                    _end_dump_file(stream)
                    stream.close()
                    file_counter += 1
                    stream = open("gcp_trivy_scanner_{}.sql".format(str(file_counter).zfill(6)), "w")
                    _start_dump_file(stream, file_counter)

                insert_str = "INSERT INTO `{}` ({}) VALUES ".format(table, ",".join(f[1] for f in fields))
                packet_length = len(insert_str)
                stream.write(insert_str)
                if not first_row:
                    row_str = row_str[1:]
            
            stream.write(row_str)
            packet_length += len(row_str)
            first_row = False

            row = c.fetchone()
            row_counter += 1

        if not empty_table:
            stream.write(";\n")

        stream.write(f"/*!40000 ALTER TABLE `{table}` ENABLE KEYS */;\n"+
                    "UNLOCK TABLES;\n\n")
        
        _end_dump_file(stream)
        stream.close()

    _close_cursor(c)

except:
    traceback.print_exc()
finally:
    _close_connection(conn)


import sys
import json
import logging
import os
import tempfile
import threading
import codecs
import traceback
import time
from queue import Queue

from enum import Enum
from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps
from dust.entity import Store, Entity, get_unit_deps_tuple

#PATH = "/var/local/beaconing/messagequeue"

UNIT_MESSAGES = "messages"
UNIT_MESSAGES_META = "messages_meta"
UNIT_ID = 5
UNIT_META_ID = 6

def get_unit_dependencies():
    return [
        get_unit_deps_tuple("dust.events", "UNIT_EVENTS", "EventTypes")
    ]

class MessageType(Enum):
    ENTITY_ACCESS = 0

class MessageMeta(MetaProps):
    message_type = (Datatypes.STRING, ValueTypes.SINGLE, 1, 100)
    message_params = (Datatypes.JSON, ValueTypes.MAP, 2, 101)
    datetime = (Datatypes.ENTITY, ValueTypes.SINGLE, 3, 102)
    entities = (Datatypes.ENTITY, ValueTypes.SET, 4, 103)
    callback_name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 104)

class MessageQueueInfoMeta(MetaProps):
    chunksize = (Datatypes.INT, ValueTypes.SINGLE, 1, 200)
    size = (Datatypes.INT, ValueTypes.SINGLE, 2, 201)
    tail = (Datatypes.INT, ValueTypes.LIST, 3, 202)
    head = (Datatypes.INT, ValueTypes.LIST, 4, 203)

class MessageTypes(FieldProps):
    message = (UNIT_MESSAGES_META, MessageMeta, 1)
    message_queue_info = (UNIT_MESSAGES_META, MessageQueueInfoMeta, 2)

Store.create_unit(UNIT_MESSAGES, UNIT_ID)
Store.load_types_from_enum(MessageTypes, UNIT_META_ID)

_log = logging.getLogger(__name__)
_log.setLevel(logging.DEBUG)
handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.DEBUG)
_log.addHandler(handler)


class EmptyMessageQueue(Exception):
    pass

class FullMessageQueue(Exception):
    pass
'''
def _truncate(fn, length):
    with open(fn, 'ab+') as f:
        f.truncate(length)

def atomic_rename(src, dst):
    try:
        os.replace(src, dst)
    except: 
        traceback.print_exc()


class MessageQueue():
    def __init__(self, path, maxsize=0, chunksize=100, tempdir=None, serializer=EntityJsonSerializer(), autosave=False):
        """Create a persistent queue object on a given path.
        The argument path indicates a directory where enqueued data should be
        persisted. If the directory doesn't exist, one will be created.
        If maxsize is <= 0, the queue size is infinite. The optional argument
        chunksize indicates how many entries should exist in each chunk file on
        disk.
        The tempdir parameter indicates where temporary files should be stored.
        The tempdir has to be located on the same disk as the enqueued data in
        order to obtain atomic operations.

        The serializer parameter controls how enqueued data is serialized. It
        must have methods dump(value, fp) and load(fp). The dump method must
        serialize value and write it to fp, and may be called for multiple
        values with the same fp. The load method must deserialize and return
        one value from fp, and may be called multiple times with the same fp
        to read multiple values.

        The autosave parameter controls when data removed from the queue is
        persisted. By default (disabled), the change is only persisted when
        task_done() is called. If autosave is enabled, data is persisted
        immediately when get() is called. Adding data to the queue with put()
        will always persist immediately regardless of this setting.
        """
        _log.debug('Initializing File based Queue with path {}'.format(path))
        self.path = path
        self.chunksize = chunksize
        self.tempdir = tempdir
        self.maxsize = maxsize
        self.serializer = serializer
        self.autosave = autosave
        self._init(maxsize)
        if self.tempdir:
            if os.stat(self.path).st_dev != os.stat(self.tempdir).st_dev:
                raise ValueError("tempdir has to be located on same path filesystem")
        else:
            fd, tempdir = tempfile.mkstemp()
            if os.stat(self.path).st_dev != os.stat(tempdir).st_dev:
                self.tempdir = self.path
                _log.warning("Default tempdir '%(dft_dir)s' is not on the " +
                            "same filesystem with queue path '%(queue_path)s'" +
                            ",defaulting to '%(new_path)s'." % {
                                "dft_dir": tempdir,
                                "queue_path": self.path,
                                "new_path": self.tempdir})
            os.close(fd)
            os.remove(tempdir)

        self.info = self._loadinfo()
        # truncate head case it contains garbage
        hnum, hcnt, hoffset = self.info.access(Operation.GET, [0, 0, 0], MessageQueueInfoMeta.head)
        headfn = self._qfile(hnum)
        if os.path.exists(headfn):
            if hoffset < os.path.getsize(headfn):
                _truncate(headfn, hoffset)
        # let the head file open
        self.headf = self._openchunk(hnum, 'ab+')
        # let the tail file open
        tnum, _, toffset = self.info.access(Operation.GET, [0, 0, 0], MessageQueueInfoMeta.tail)
        self.tailf = self._openchunk(tnum)
        self.tailf.seek(toffset)
        # update unfinished tasks with the current number of enqueued tasks
        self.unfinished_tasks = self.info.access(Operation.GET, 0, MessageQueueInfoMeta.size)
        # optimize info file updates
        self.update_info = True

    def _init(self, maxsize):
        self.mutex = threading.Lock()
        self.not_empty = threading.Condition(self.mutex)
        self.not_full = threading.Condition(self.mutex)
        self.all_tasks_done = threading.Condition(self.mutex)

        if not os.path.exists(self.path):
            os.makedirs(self.path)

    def join(self):
        with self.all_tasks_done:
            while self.unfinished_tasks:
                self.all_tasks_done.wait()

    def qsize(self):
        n = None
        with self.mutex:
            n = self._qsize()
        return n

    def _qsize(self):
        return self.info.access(Operation.GET, None, MessageQueueInfoMeta.size)

    def empty(self):
        return self.qsize() == 0

    def put(self, item, block=True, timeout=None):
        "Interface for putting item in disk-based queue."
        self.not_full.acquire()
        try:
            if self.maxsize > 0:
                if not block:
                    if self._qsize() == self.maxsize:
                        raise FullMessageQueue
                elif timeout is None:
                    while self._qsize() == self.maxsize:
                        print("waiting for not full")
                        self.not_full.wait()
                elif timeout < 0:
                    raise ValueError("'timeout' must be a non-negative number")
                else:
                    endtime = _time() + timeout
                    while self._qsize() == self.maxsize:
                        remaining = endtime - _time()
                        if remaining <= 0.0:
                            raise FullMessageQueue
                        print("waiting for not full")
                        self.not_full.wait(remaining)
            self._put(item)
            self.unfinished_tasks += 1
            self.not_empty.notify()
        finally:
            self.not_full.release()

    def _put(self, item):
        self.serializer.dump(item, self.headf)
        self.headf.flush()
        print(str(self.info.access(Operation.GET, None, MessageQueueInfoMeta.head)))
        hnum, hpos, _ = self.info.access(Operation.GET, None, MessageQueueInfoMeta.head)
        hpos += 1
        if hpos == self.info.access(Operation.GET, None, MessageQueueInfoMeta.chunksize):
            hpos = 0
            hnum += 1
            # make sure data is written to disk whatever
            # its underlying file system
            os.fsync(self.headf.fileno())
            self.headf.close()
            self.headf = self._openchunk(hnum, 'ab+')
        self.info.access(Operation.CHANGE, 1, MessageQueueInfoMeta.size)
        self.info.access(Operation.SET, [hnum, hpos, self.headf.tell()], MessageQueueInfoMeta.head)
        self._saveinfo()

    def put_nowait(self, item):
        return self.put(item, False)

    def get(self, block=True, timeout=None):
        self.not_empty.acquire()
        try:
            if not block:
                if not self._qsize():
                    raise EmptyMessageQueue
            elif timeout is None:
                while not self._qsize():
                    self.not_empty.wait()
            elif timeout < 0:
                raise ValueError("'timeout' must be a non-negative number")
            else:
                endtime = _time() + timeout
                while not self._qsize():
                    remaining = endtime - _time()
                    if remaining <= 0.0:
                        raise EmptyMessageQueue
                    self.not_empty.wait(remaining)
            item = self._get()
            self.not_full.notify()
            return item
        finally:
            self.not_empty.release()

    def get_nowait(self):
        return self.get(False)

    def _get(self):
        tnum, tcnt, toffset = self.info.access(Operation.GET, None, MessageQueueInfoMeta.tail)
        hnum, hcnt, _ = self.info.access(Operation.GET, None, MessageQueueInfoMeta.head)
        if [tnum, tcnt] >= [hnum, hcnt]:
            return None
        data = self.serializer.load(self.tailf)
        toffset = self.tailf.tell()
        tcnt += 1
        if tcnt == self.info.access(Operation.GET, None, MessageQueueInfoMeta.chunksize) and tnum <= hnum:
            tcnt = toffset = 0
            tnum += 1
            self.tailf.close()
            self.tailf = self._openchunk(tnum)
        self.info.access(Operation.CHANGE, -1, MessageQueueInfoMeta.size)
        self.info.access(Operation.SET, [tnum, tcnt, toffset], MessageQueueInfoMeta.tail)
        if self.autosave:
            self._saveinfo()
            self.update_info = False
        else:
            self.update_info = True
        return data

    def task_done(self):
        with self.all_tasks_done:
            unfinished = self.unfinished_tasks - 1
            if unfinished <= 0:
                if unfinished < 0:
                    raise ValueError("task_done() called too many times.")
                self.all_tasks_done.notify_all()
            self.unfinished_tasks = unfinished
            self._task_done()

    def _task_done(self):
        if self.autosave:
            return
        if self.update_info:
            self._saveinfo()
            self.update_info = False

    def _openchunk(self, number, mode='rb'):
        return open(self._qfile(number), mode)

    def _loadinfo(self):
        infopath = self._infopath()
        if os.path.exists(infopath):
            with open(infopath, 'rb') as f:
                info = self.serializer.load(f)
                return info[0]
        else:
            info = Store.access(Operation.GET, None, UNIT_MESSAGES, 1, MessageTypes.message_queue_info)
            info.access(Operation.SET, self.chunksize, MessageQueueInfoMeta.chunksize)
            info.access(Operation.SET, 0, MessageQueueInfoMeta.size)
            info.access(Operation.SET, [0, 0, 0], MessageQueueInfoMeta.tail)
            info.access(Operation.SET, [0, 0, 0], MessageQueueInfoMeta.head)

            print("INfo is {}".format(info.access(Operation.GET, None, MessageQueueInfoMeta.head)))

        return info

    def _gettempfile(self):
        if self.tempdir:
            return tempfile.mkstemp(dir=self.tempdir)
        else:
            return tempfile.mkstemp()

    def _saveinfo(self):
        tmpfd, tmpfn = self._gettempfile()
        with os.fdopen(tmpfd, "wb") as tmpfo:
            self.serializer.dump(self.info, tmpfo)
        atomic_rename(tmpfn, self._infopath())
        self._clear_tail_file()

    def _clear_tail_file(self):
        """Remove the tail files whose items were already get."""
        tnum, _, _ = self.info.access(Operation.GET, None, MessageQueueInfoMeta.tail)
        while tnum >= 1:
            tnum -= 1
            path = self._qfile(tnum)
            if os.path.exists(path):
                os.remove(path)
            else:
                break

    def _qfile(self, number):
        return os.path.join(self.path, 'q%05d' % number)

    def _infopath(self):
        return os.path.join(self.path, 'info')

    def __del__(self):
        """Handles the removal of queue."""
        for to_close in [self.headf, self.tailf]:
            if to_close and not to_close.closed:
                to_close.close() 
'''
def register_listener(name, entity_filter, cb):
    _listeners[name] = (entity_filter, cb)


def unregister_listener(name):
    if name in _listeners:
        del _listeners[name]

_stop = False
_queue = Queue()
_listeners = {}

def signal_finish():
    global _stop

    _queue.join()
    _stop = True
    _queue.put(Store.access(Operation.GET, None, UNIT_MESSAGES, None, MessageTypes.message))

def start_queue_processor(queue, log):
    global _stop

    while True:
        try:
            entity = _queue.get()
            if entity == None:
                time.sleep(0.5)
            else:
                _log.debug("Processing item: {}".format(entity.global_id()))
                try:
                    _listeners[entity.access(Operation.GET, None, MessageMeta.callback_name)][1](
                        entity.access(Operation.GET, None, MessageMeta.message_type),
                        entity.access(Operation.GET, None, MessageMeta.message_params),
                        entity.access(Operation.GET, None, MessageMeta.entities)
                    )
                except KeyError:
                    if entity.access(Operation.GET, None, MessageMeta.callback_name):
                        _log.error("Invalid callback registered: {}".format(entity.access(Operation.GET, None, MessageMeta.callback_name)))
            if _stop:
                break

        except EmptyMessageQueue:
            time.sleep(0.5)
            if _stop:
                break
        except KeyboardInterrupt:
            _log.warning("Keyboard interrupt received")
            break
        except:
            traceback.print_exc()
        finally:
            _queue.task_done()

def create_message(message_type, message_params, entities):
    for callback_name, listener in _listeners.items():
        entity_filter, cb = listener
        if entity_filter(message_type, message_params, entities):
            message = Store.access(Operation.GET, None, UNIT_MESSAGES, None, MessageTypes.message)
            message.access(Operation.SET, message_type.name, MessageMeta.message_type)
            message.access(Operation.SET, callback_name, MessageMeta.callback_name)
            if message_params:
                message.access(Operation.SET, message_params, MessageMeta.message_params)
            if entities:
                for e in entities:
                    if e:
                        message.access(Operation.ADD, entities, MessageMeta.entities)
            _queue.put(message)


_queue_processor = threading.Thread(target=start_queue_processor, args=(_queue, _log, ), daemon=False)
_queue_processor.start()


import json
import traceback
import calendar
import pytz
import datetime

from enum import Enum
from dateutil import parser

from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps
from dust.entity import Store

UNIT_EVENTS = "events"
UNIT_EVENTS_META = "events_meta"
UNIT_ID = 3
UNIT_META_ID = 4

UTC = pytz.timezone('UTC')

FORMAT_DATETIME_EU = "%d.%m.%Y %H:%M:%S"
FORMAT_DATETIME_SHORT_EU = "%d.%m.%y %H:%M"
FORMAT_DATETIME_EU_TZ = "%d.%m.%Y %H:%M:%S %Z"

FORMAT_DATE_EU = "%d.%m.%Y"
FORMAT_DB_DATE = "%Y-%m-%d"
FORMAT_DB_DATETIME = "%Y-%m-%d %H:%M:%S"
FORMAT_COMPRESSED_DATETIME = "%Y%m%d%H%M%S"

class RepeatTypes(Enum):
    NO_REPEAT = 0
    DAILY = 1
    WEEKLY = 2
    MONTHLY = 3
    BYWEEKLY = 4
    YEARLY = 5
    DAYOFWEEK = 6
    CUSTOM = 7

class EventType(Enum):
    DATE = 0
    TIME = 1
    DATETIME = 2

class EventMeta(MetaProps):
    start = (Datatypes.INT, ValueTypes.SINGLE, 1, 100)
    duration_in_sec = (Datatypes.INT, ValueTypes.SINGLE, 2, 101)
    repeat = (Datatypes.STRING, ValueTypes.SINGLE, 3, 102)
    repeat_value = (Datatypes.INT, ValueTypes.LIST, 4, 103)
    repeat_until = (Datatypes.INT, ValueTypes.LIST, 5, 104)

class EventTypes(FieldProps):
    event = (UNIT_EVENTS_META, EventMeta, 1)

Store.create_unit(UNIT_EVENTS, UNIT_ID)
Store.load_types_from_enum(EventTypes, UNIT_META_ID)

def parse_event(event_value_start, event_type, iso=False, duration_in_sec=None, repeat_type=RepeatTypes.NO_REPEAT, repeat_value=None, repeat_until=None, ignoretz=False, tzinfos=None, tz=None):
    try:
        if iso:
            dt = parser.isoparse(event_value_start)
        else:
            dt = parser.parse(event_value_start, ignoretz=ignoretz, tzinfos=tzinfos)
        return get_event(dt, event_type, duration_in_sec, repeat_type, repeat_value, repeat_until, tz)
    except:
        traceback.print_exc()

    return None

def get_event(dt, event_type, duration_in_sec=None, repeat_type=RepeatTypes.NO_REPEAT, repeat_value=None, repeat_until=None, tz=None):
    event = Store.access(Operation.GET, None, UNIT_EVENTS, None, EventTypes.event)
    if not tz is None:
        if _is_naive(dt):
            dt = tz.localize(dt, is_dst=None)
        else:
            dt = dt.astimezone(tz)
    else:
        if _is_naive(dt):
            dt = UTC.localize(dt, is_dst=None)

    event.access(Operation.SET, int(dt.timestamp()), EventMeta.start)
    event.access(Operation.SET, duration_in_sec, EventMeta.duration_in_sec)
    event.access(Operation.SET, repeat_type.name, EventMeta.repeat)

    return event

def format_event(event, format_string=FORMAT_DATETIME_EU, tz=UTC):
    repeat = RepeatTypes[event.access(Operation.GET, None, EventMeta.repeat)]
    if repeat == RepeatTypes.NO_REPEAT:
        dt = datetime.datetime.fromtimestamp(event.access(Operation.GET, None, EventMeta.start), tz)
        return datetime.datetime.strftime(dt, format_string)

    return ""


def _is_naive(dt):
    return dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None



from dust import logger
from dust.httpservices import SVCINFO_USER, SVCINFO_AUTHINFO, SVCINFO_CLIENTIP, SVCINFO_TIMESTAMP, SVCLOG_RESPONSE, SVCRESP_STATUS
from dust.httpservices.actioncontroller import ActionController

from datetime import datetime
import traceback
import json

class CommandHandler():
    def __init__(self, server, request_wrapper, *service_names):
        self.server = server
        self.request_wrapper = request_wrapper
        self.action_ctrl = ActionController(*service_names)

    def send_json_response(self, esp_data, http_status):
        pass

    def get_session_user(self, request):
        pass

    def handle(self, target, request, response):
        request = self.request_wrapper(request)
        cmd = target

        if cmd.startswith("/"):
            cmd = cmd[1:]

        param = None
        param_idx = cmd.find("/")

        if param_idx != -1:
            param = cmd[param_idx+1:]
            cmd = cmd[0:param_idx]

        logger().debug("Server cmd: {}, param: {}".format(cmd, param))

        success = False
        resp_data = {}
        http_status = 204

        try:
            if cmd == "getSessionUser":
                resp_data = self.get_session_user(request)

            elif cmd == "api":
                input_data = request.get_json()
                if input_data == None:
                    input_data = {}

                input_data.update(request.get_query_params())

                self.log(input_data, "Input data is:")

                auth_data = self.get_session_user(request)

                input_data[SVCINFO_USER] = auth_data.get("uid")
                input_data[SVCINFO_AUTHINFO] = auth_data
                input_data[SVCINFO_CLIENTIP] = request.remote_addr()
                input_data[SVCINFO_TIMESTAMP] = datetime.now()

                success = self.action_ctrl.relay_action(input_data, request, resp_data)
                if success:
                    http_status = 200

                pass 
            elif cmd == "ping":
                pass 

        except:
            traceback.print_exc()

        if resp_data:
            response[SVCRESP_STATUS] = success
            response[SVCLOG_RESPONSE] = resp_data


    def log(self, json_data, msg=None):
        if msg:
            logger().debug(msg)
        logger().debug(json.dumps(json_data, indent=4))


import mysql.connector
from mysql.connector import errorcode
import traceback
import os

from dust.persist.sqlpersist import SqlPersist

from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps
from dust.entity import Entity

SQL_TYPE_MAP = {
    Datatypes.INT: "INT",
    Datatypes.NUMERIC: "DOUBLE",
    Datatypes.BOOL: "TINYINT",
    Datatypes.STRING: "TEXT",
    Datatypes.BYTES: "MEDIUMBLOB",
    Datatypes.JSON: "TEXT",
    Datatypes.ENTITY: "TEXT"
}

CREATE_TABLE_TEMPLATE = "\
CREATE TABLE IF NOT EXISTS {{sql_table.table_name}} (\n\
    {% for field in sql_table.fields %}\
    {{ field.field_name }} {{ field.field_type }}{% if field.primary_key %} PRIMARY KEY{% endif %}{% if not loop.last %},{% endif %}\n\
    {% endfor %}\
)\n\
"

CREATE_TABLE_TEMPLATE_MULTI_PK = "\
CREATE TABLE IF NOT EXISTS {{sql_table.table_name}} (\n\
    {% for field in sql_table.fields %}\
    {{ field.field_name }} {{ field.field_type }},\n\
    {% endfor %}\n\
    PRIMARY KEY ({% for field in sql_table.primary_keys %}{{ field.field_name }}{% if not loop.last %},{% endif %}{% endfor %})\n\
)\n\
"

ALTER_TABLE_TEMPLATE = "\
ALTER TABLE {{sql_table.table_name}} \n\
    {% for field_mod in field_modifications %}\
    {{ field_mod }}{% if not loop.last %},{% endif %}\n\
    {% endfor %}\
\n\
"

DROP_TABLE_TEMPLATE = "\
DROP TABLE {{sql_table.table_name}}"

'''
INSERT_INTO_TABLE_TEMPLATE = "\
INSERT INTO {{sql_table.table_name}} (\
{% for field in sql_table.fields %}\
{{ field.field_name }}{% if not loop.last %},{% endif %}\
{% endfor %}\
) VALUES (\
{% for field in sql_table.fields %}\
%({{ field.field_name }})s{% if not loop.last %},{% endif %}\
{% endfor %}\
)\
"
'''
INSERT_INTO_TABLE_TEMPLATE = "\
INSERT INTO {{sql_table.table_name}} (\
{% for field in sql_table.fields %}\
{{ field.field_name }}{% if not loop.last %},{% endif %}\
{% endfor %}\
) VALUES (\
{% for field in sql_table.fields %}\
%s{% if not loop.last %},{% endif %}\
{% endfor %}\
)\
"


SELECT_TEMPLATE = "\
SELECT \
{% for field in sql_table.fields %}\
{{ field.field_name }}{% if not loop.last %},{% endif %} \
{% endfor %}\
FROM {{sql_table.table_name}} \
{% if where_filters %}\
WHERE \
{% for filter in where_filters %}\
{% if filter[1]|lower == 'in' %}\
{{ filter[0] }} {{ filter[1] }} ({% for v in filter[2] %}%({{ filter[0] }}{{loop.index}})s{% if not loop.last %},{% endif %}{% endfor %}) {% if not loop.last %}AND {% endif %}\
{% else %}\
{{ filter[0] }} {{ filter[1] }} %({{ filter[0] }})s {% if not loop.last %}AND {% endif %}\
{% endif %}\
{% endfor %}\
{% endif %}\
"

'''
UPDATE_TEMPLATE = "\
UPDATE {{sql_table.table_name}} SET \
{% for field in sql_table.fields %}\
{% if not field.primary_key and not field.base_field %}{{ field.field_name }} = %({{ field.field_name }})s{% if not loop.last %},{% endif %}{% endif %} \
{% endfor %}\
WHERE \
{% for field in sql_table.primary_keys %}{{ field.field_name }} = %({{ field.field_name }})s{% if not loop.last %},{% endif %}{% endfor %} \
"
'''
UPDATE_TEMPLATE = "\
UPDATE {{sql_table.table_name}} SET \
{% for field in sql_table.fields %}\
{% if not field.primary_key and not field.base_field %}{{ field.field_name }} = %s{% if not loop.last %},{% endif %}{% endif %} \
{% endfor %}\
WHERE \
{% for field in sql_table.primary_keys %}{{ field.field_name }} = %s{% if not loop.last %},{% endif %}{% endfor %} \
"

DELETE_TEMPLATE = "\
DELETE FROM {{sql_table.table_name}} \
WHERE _global_id = %s\
"

DELETE_TEMPLATE_WITH_WHERE = "\
DELETE FROM {{sql_table}} \
WHERE \
{% for where_filter in where_filters %}{{ where_filter[0] }} {{ where_filter[1] }} %({{ where_filter[0] }})s{% if not loop.last %} and {% endif %}{% endfor %} \
"

MYSQL_USER = os.environ.get('MYSQL_USER')
MYSQL_PASSWORD = os.environ.get('MYSQL_PASSWORD')
MYSQL_HOST = os.environ.get('MYSQL_HOST')
MYSQL_DB = os.environ.get('MYSQL_DB')

class MySQLPersist(SqlPersist):
    def __init__(self):
        super().__init__(**self.__db_api_kwargs())

    def __create_connection(self):
        conn = None
        try:
            conn = mysql.connector.connect(user=MYSQL_USER, password=MYSQL_PASSWORD, host=MYSQL_HOST, database=MYSQL_DB, charset='utf8', use_unicode=True)
        except mysql.connector.Error as err:
            if err.errno == errorcode.ER_ACCESS_DENIED_ERROR:
                print("Something is wrong with your user name or password")
            elif err.errno == errorcode.ER_BAD_DB_ERROR:
                print("Database does not exist")
            else:
                print(err)

        return conn

    def __close_connection(self, conn):
        if conn:
            conn.commit()
            conn.close()

    def __create_cursor(self, conn, prepared=False, buffered=True):
        if conn:
            return conn.cursor(buffered=buffered, prepared=prepared)

    def __close_cursor(self, c):
        if c:
            c.close()

    def __db_api_kwargs(self):
        return {
            "_create_connection": self.__create_connection,
            "_close_connection": self.__close_connection,
            "_create_cursor": self.__create_cursor,
            "_close_cursor": self.__close_cursor
        }

    def create_exectute_params(self, named_param=True):
        if named_param:
            return {}
        else:
            return []

    def add_execute_param(self, values, name, value, operator="=", named_param=True):
        if named_param:
            if operator and operator.lower() == "in":
                cnt = 1
                for v in value:
                    values[name+str(cnt)] = v
                    cnt += 1
            else:
                values[name] = value
        else:
            if operator and operator.lower() == "in":
                cnt = 1
                for v in value:
                    values.append(v)
            else:
                values.append(value)

    def table_exits(self, table_name, conn):
        try:
            cursor = self.__create_cursor(conn)
            cursor.execute("SELECT * FROM information_schema.tables WHERE table_schema = %(table_schema)s AND table_name = %(table_name)s LIMIT 1;", {"table_schema": MYSQL_DB, "table_name": table_name})
            rows = cursor.fetchall()

            for row in rows:
                if row[2] == table_name:
                    return True
        except:
            traceback.print_exc()
        finally:
            self.__close_cursor(cursor)

        return False

    def create_table_template(self, sql_table):
        if len(sql_table.primary_keys) > 1:
            return CREATE_TABLE_TEMPLATE_MULTI_PK
        else:
            return CREATE_TABLE_TEMPLATE 

    def alter_table_template(self):
        return ALTER_TABLE_TEMPLATE 

    def delete_table_template(self):
        return DROP_TABLE_TEMPLATE 

    def alter_table(self, sql, conn):
        cursor = None
        try:
            cursor = self.__create_cursor(conn)
            cursor.execute(sql)
        except:
            print(sql)
            traceback.print_exc()
        finally:
            self.__close_cursor(cursor)

    def insert_into_table_template(self):
        return INSERT_INTO_TABLE_TEMPLATE

    def select_template(self, where_filters):
        return SELECT_TEMPLATE

    def update_template(self):
        return UPDATE_TEMPLATE

    def delete_template(self, where_filters=None):
        if where_filters:
            return DELETE_TEMPLATE_WITH_WHERE
        else:
            return DELETE_TEMPLATE

    def convert_value_to_db(self, field, value):
        if field.datatype == Datatypes.BOOL:
            if value == True:
                return 1
            else:
                return 0
        elif field.datatype == Datatypes.ENTITY and isinstance(value, Entity):
            return value.global_id()
        elif SQL_TYPE_MAP[field.datatype] == "TEXT":
            if value is not None and len(value) > 65500:
                print(f"Trunctating value to 65500 for {field.name}")
                return value[:65500]
            else:
                return value
        elif SQL_TYPE_MAP[field.datatype] == "MEDIUMBLOB":
            if value is not None and len(value) > 16777000:
                print(f"Value is longer than max size, returning None for {field.name}")
                return None
            else:
                return value
        else:
            return value

    def convert_value_from_db(self, field, value):
        if field.datatype == Datatypes.BOOL:
            if value == 1:
                return True
            else:
                return False
        else:
            return value

    def sql_type(self, datatype, valuetype, primary_key=False):
        if primary_key and datatype == Datatypes.STRING:
            return "VARCHAR(100)"
        elif valuetype:
            return SQL_TYPE_MAP[datatype]
        else:
            return "TEXT"



from collections import defaultdict

# Chart type registry
CHART_HANDLERS = {}

def register_chart_handler(chart_type):
    def decorator(fn):
        CHART_HANDLERS[chart_type] = fn
        return fn
    return decorator

def generate_chart_data(cube, config):
    chart_type = config.get("type")
    if chart_type not in CHART_HANDLERS:
        raise ValueError(f"Unsupported chart type: {chart_type}")
    return CHART_HANDLERS[chart_type](cube, config)


@register_chart_handler("bar")
def generate_bar_chart(cube, config):
    # Extract config callables
    visit_fn = config["visit"]
    x_fn = config["x"]
    count_fn = config["count"]
    filter_fn = config.get("filter", lambda *_: True)
    stack_fn = config.get("stack_by")
    colors = config.get("colors")
    label_metric_fn = config.get("label_metric")
    label_formatter_fn = config.get("label_formatter")

    x_order = config.get("x_order")
    stack_order = config.get("stack_order")

    # Aggregated chart data
    bar_data = defaultdict(lambda: defaultdict(int))
    label_tracker = defaultdict(set)  # To track label metrics per x

    for coords, value in visit_fn(cube):
        if not filter_fn(coords, value):
            continue

        x_key = x_fn(coords, value)
        stack_key = stack_fn(coords, value) if stack_fn else None
        count = count_fn(coords, value)

        # Aggregate count
        if stack_fn:
            bar_data[x_key][stack_key] += count
        else:
            bar_data[x_key] += count

        # Track label metric if provided
        if label_metric_fn:
            label_tracker[x_key].add(label_metric_fn(coords, value))

    # Optional per-bar annotations
    x_labels = x_order or list(bar_data.keys())
    annotations = {
        x: label_formatter_fn(label_tracker[x])
        for x in x_labels
    } if label_metric_fn and label_formatter_fn else {}

    if stack_fn is None:
        # Not a stacked chart, just one dataset
        dataset = {
            "label": config.get("label", "value"),  # optional override
            "data": [],
            "backgroundColor": colors.get(config.get("label", "value"))
        }

        # Prepare ordered labels
        for x in x_labels:
            dataset["data"].append(bar_data[x])  # value is directly int

        return {
            "type": "bar",
            "data": {
                "labels": x_labels,
                "datasets": [dataset],
            },
            "annotations": annotations,
            "options": config.get("options", {})
        }
    else:
        # Prepare ordered labels
        stacks = stack_order or sorted({s for bars in bar_data.values() for s in bars})
        # Prepare datasets
        datasets = []
        for stack in stacks:
            datasets.append({
                "label": stack,
                "data": [bar_data[x].get(stack, 0) for x in x_labels],
                "backgroundColor": colors.get(stack)
            })


        return {
            "type": "bar",
            "data": {
                "labels": x_labels,
                "datasets": datasets,
            },
            "annotations": annotations,
            "options": config.get("options", {})
        }
    
@register_chart_handler("line")
def generate_line_chart(cube, config):
    from collections import defaultdict

    # Required config callbacks
    visit_fn = config["visit"]             # Function returning iterator of (coords, value)
    x_fn = config["x"]                     # Extracts x-axis label (e.g., period)
    y_fn = config["y"]                     # Extracts y-axis value (e.g., count)

    # Optional config callbacks
    filter_fn = config.get("filter", lambda *_: True)  # Optional filter on (coords, value)
    series_fn = config.get("series_by")                # Optional function to split data into series
    label_formatter_fn = config.get("label_formatter") # Optional formatting of collected labels

    # Optional config parameters
    colors = config.get("colors", {})                  # Color per series (dict)
    x_order = config.get("x_order")                    # Optional predefined order of x-axis labels
    series_order = config.get("series_order")          # Optional predefined order of series keys

    DEFAULT_SERIES_KEY = "__default__"

    # Store data points: series_key -> x_key -> y_value
    line_data = defaultdict(lambda: defaultdict(float))

    # Collect label metrics per x_key (for annotation)
    label_tracker = {}

    x_keys_seen = []  # Track x_keys in visit order without duplication 

    # Pass 1: visit cube, aggregate values and collect annotation metrics
    for coords, value in visit_fn(cube):
        if not filter_fn(coords, value):
            continue

        x_key = x_fn(coords, value) 
        y_value = y_fn(coords, value)  # e.g., count

        # Track x_key insertion order (preserve natural iteration order)
        if x_key not in x_keys_seen:
            x_keys_seen.append(x_key)

        series_keys = series_fn(coords, value) if series_fn else DEFAULT_SERIES_KEY

        # Ensure we always work with a list of series keys
        if not isinstance(series_keys, list):
            series_keys = [series_keys]

        for key in series_keys:
            line_data[key][x_key] += y_value

    # Sort axis labels and series keys
    x_labels = x_order or x_keys_seen
    if label_formatter_fn:
        for x_label in x_labels:
            label_tracker[x_label] = label_formatter_fn(x_label)

    series_order_keys = series_order or sorted(line_data.keys())

    # Build annotation labels using formatter function
    annotations = {
        x: label_tracker[x]
        for x in x_labels
    } if label_formatter_fn else {}

    # Build chart.js dataset entries
    datasets = []
    for series in series_order_keys:
        datasets.append({
            "label": None if series == DEFAULT_SERIES_KEY else series,
            "data": [line_data[series].get(x, 0) for x in x_labels],
            "borderColor": colors.get(series, "#000000"),
            "backgroundColor": colors.get(series, "#000000"),
            "fill": True,
            "tension": 0.1  # Smooth out the line slightly
        })

    # Return final chart config for chart.js
    return {
        "type": "line",
        "data": {
            "labels": x_labels,
            "datasets": datasets
        },
        "annotations": annotations,
        "options": config.get("options", {})
    }



from dust.httpservices import DustResultType
import gzip
import json
import base64

import traceback

_services = {}

class ServiceBase():
	def __init__(self, module):
		self.module = module
		_services[module] = self

	@staticmethod
	def get_service(name):
		return _services.get(name)

	def do_process(self, params, request, response, immediate=True):
		if immediate:
			return DustResultType.ACCEPT
		else:
			return DustResultType.NOTIMPLEMENTED

	def get_modulename(self):
		return self.module 

	def compress(self, filename, json_data):
		encoded_data = None
		try:
			with gzip.open(filename, 'w') as fout:
				fout.write(json.dumps(json_data).encode('utf-8'))  

			with gzip.open(filename, 'r') as fin:
				encoded_data = base64.b64encode(fin.read()).decode()
		except:
			traceback.print_exc()

		return encoded_data



from enum import Enum

__all__ = ["actioncontroller","servicebase","commandhandler"]

SVCINFO_USER = "userId"
SVCINFO_AUTHINFO = "AuthInfo"
SVCINFO_CLIENTIP = "clientIp"
SVCINFO_TIMESTAMP = "timestamp"
SVCINFO_MODULE = "module"

SVCPARAM_COMMAND = "command"
SVCPARAM_TARGET = "target"
SVCPARAM_CONTENT = "Content"
SVCPARAM_EXTRAPARAMS = "ExtraParams"

SVCRESP_STATUS = "status"
SVCRESP_ERRORMSG = "errorMsg"
SVCLOG_EXECLOG = "ExecLog"

SVCLOG_REQUEST = "Request"
SVCLOG_RESPONSE = "Response"
    
SVCREJECT_STATUS = "httpStatus"
SVCREJECT_CHECK = "rejectCheck"
SVCREJECT_MESSAGE = "rejectMessage"


class DustResultType(Enum):
    NOTIMPLEMENTED = 1
    REJECT = 2
    ACCEPT_PASS = 3
    ACCEPT = 4
    ACCEPT_READ = 5
    READ = 6

    def is_readon(self):
        return self.value == DustResultType.READ or self.value == DustResultType.ACCEPT_READ

    def is_reject(self):
        return self.value == DustResultType.NOTIMPLEMENTED or self.value == DustResultType.REJECT

class RequestWrapper():
    def __init__(self, request):
        self.request = request

    def get_method(self):
        pass

    def get_query_params(self):
        pass 

    def get_json(self):
        pass

    def get_path(self):
        pass

    def remote_addr(self):
        pass

    def get_headers(self):
        pass


import traceback
import requests
import json
import os

from datetime import datetime
from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps
from dust.entity import Store, Entity, _entity_map, get_unit_deps_tuple
from dust.events import get_event, EventType
from dust.messages import register_listener, unregister_listener

UNIT_SLACK = "slack"
UNIT_SLACK_META = "slack_meta"
UNIT_ID = 7
UNIT_META_ID = 8

TEST_SLACK = True

def get_unit_dependencies():
    return [
        get_unit_deps_tuple("dust.events", "UNIT_EVENTS", "EventTypes")
    ]

class SlackChannelMeta(MetaProps):
    name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 100)
    webhook_url = (Datatypes.STRING, ValueTypes.SINGLE, 2, 101)
    test_webhook_url = (Datatypes.STRING, ValueTypes.SINGLE, 3, 102)

class SlackNotificationMeta(MetaProps):
    notification = (Datatypes.STRING, ValueTypes.SINGLE, 1, 200)
    channel = (Datatypes.ENTITY, ValueTypes.SINGLE, 2, 201)
    report_time = (Datatypes.ENTITY, ValueTypes.SINGLE, 3, 202)
    sent = (Datatypes.BOOL, ValueTypes.SINGLE, 4, 203)
    test = (Datatypes.BOOL, ValueTypes.SINGLE, 4, 204)

class SlackTypes(FieldProps):
    channel = (UNIT_SLACK_META, SlackChannelMeta, 1)
    notification = (UNIT_SLACK_META, SlackNotificationMeta, 2)

def set_slack_test():
    global TEST_SLACK
    try:
        default_branch = os.environ.get('CI_DEFAULT_BRANCH')
        commit_ref_name = os.environ.get('CI_COMMIT_REF_NAME')
        pipeline_src = os.environ.get("CI_PIPELINE_SOURCE")
        if pipeline_src == "schedule" and commit_ref_name == default_branch and os.environ.get('slack_test_channel') != 'True':
            TEST_SLACK = False
        else:
            TEST_SLACK = True
    except:
        TEST_SLACK = True

Store.create_unit(UNIT_SLACK, UNIT_ID)
Store.load_types_from_enum(SlackTypes, UNIT_META_ID)
set_slack_test()

_channels = {}

def _default_event_filter(message_type, message_params, entities):
    for global_id in entities:
        entity = Store.access(Operation.GET, None, global_id)

        if message_params["op"] == Operation.SET.name and \
           entity.get_meta_type_enum() == SlackTypes.notification and \
           not entity.access(Operation.GET, None, SlackNotificationMeta.notification) is None and \
           not entity.access(Operation.GET, None, SlackNotificationMeta.channel) is None and \
           not entity.access(Operation.GET, None, SlackNotificationMeta.report_time) is None and \
           not entity.access(Operation.GET, None, SlackNotificationMeta.test) is None and \
           not entity.access(Operation.GET, False, SlackNotificationMeta.sent):
            return True

    return False


def register_slack_listener(channel_name, event_filter=_default_event_filter):
    register_listener("{}:{}".format(UNIT_SLACK, channel_name), event_filter, _notify_slack)

def unregister_slack_listener(channel_name):
    unregister_listener("{}:{}".format(UNIT_SLACK, channel_name))

def register_channel(channel_name, webhook_url, test_webhook_url=None, register_default_listener=True):
    if test_webhook_url is None:
        test_webhook_url = webhook_url

    channel = Store.access(Operation.GET, None, UNIT_SLACK, None, SlackTypes.channel)
    channel.access(Operation.SET, channel_name, SlackChannelMeta.name)
    channel.access(Operation.SET, webhook_url, SlackChannelMeta.webhook_url)
    channel.access(Operation.SET, test_webhook_url, SlackChannelMeta.test_webhook_url)

    _channels[channel_name] = channel
    if register_default_listener:
        register_slack_listener(channel_name)


def _notify_slack(message_type, message_params, entities):
    print(str(entities))
    for global_id in entities:
        entity = Store.access(Operation.GET, None, global_id)
        notification = entity.access(Operation.GET, None, SlackNotificationMeta.notification)

        if notification and len(notification.strip()) > 0:
            webhook_url = None
            if entity.access(Operation.GET, True, SlackNotificationMeta.test):
                webhook_url = entity.access(Operation.GET, None, SlackNotificationMeta.channel, SlackChannelMeta.test_webhook_url)
            else:
                webhook_url = entity.access(Operation.GET, None, SlackNotificationMeta.channel, SlackChannelMeta.webhook_url)

            print("Sending to slack")
            try:
                data = json.dumps(json.loads(notification))
            except:
                traceback.print_exc()
                data = json.dumps({'text': notification})
            response = requests.post(
                webhook_url, data = data,
                headers={'Content-Type': 'application/json'}
                )
            if response.status_code != 200:
                raise ValueError(
                    'Request to slack returned an error %s, the response is:\n%s'
                    % (response.status_code, response.text)
                    )

def create_slack_notification(channel_name, notification, test=True):
    if notification and len(notification.strip()) > 0:
        notif_entity = Store.access(Operation.GET, None, UNIT_SLACK, None, SlackTypes.notification)
        notif_entity.access(Operation.SET, notification, SlackNotificationMeta.notification)
        notif_entity.access(Operation.SET, _channels[channel_name], SlackNotificationMeta.channel)
        notif_entity.access(Operation.SET, get_event(datetime.now(), EventType.DATETIME), SlackNotificationMeta.report_time)
        notif_entity.access(Operation.SET, False, SlackNotificationMeta.sent)
        notif_entity.access(Operation.SET, test, SlackNotificationMeta.test)

        return notif_entity

    return None




"""
DustStat: A multi-dimensional data cube implementation in Python.

This module provides a way to store and manipulate values in an n-dimensional space,
where each dimension is defined by an axis. Data is stored in a dictionary, indexed
by coordinates (tuples of indices corresponding to axis categories). The cube supports
retrieving, setting, and incrementing values at specific coordinates.
"""

import json
import yaml
import hashlib
from collections import defaultdict
from typing import List, Dict, Tuple, Union, Optional, Callable, Generator
from enum import Enum
from abc import ABC, abstractmethod

class SortOrder(Enum):
    ASCENDING = "asc"
    DESCENDING = "desc"

class DustStatCubeType(Enum):
    STRUCTURED = "structured"

class DustStatAxis(ABC):
    """Base class for an axis in the data cube."""
    
    def __init__(self, name: str):
        self.name: str = name
        self.metadata: Dict[str, Dict[str, Union[str, int, float]]] = {}

    @abstractmethod
    def categories(self) -> List[str]:
        """Must be implemented in derived classes to return the categories of the axis."""
        pass

    def set_metadata(self, category: str, key: str, value: Union[str, int, float]) -> None:
        """Stores metadata for a given category in this axis."""
        if category not in self.metadata:
            self.metadata[category] = {}
        self.metadata[category][key] = value

    def get_metadata(self, category: str, key: str) -> Optional[Union[str, int, float]]:
        """Retrieves metadata for a given category."""
        return self.metadata.get(category, {}).get(key, None)


class DustStatAxisConst(DustStatAxis):
    def __init__(self, name: str, categories: Optional[List[str]] = None):
        super().__init__(name)
        self._categories: List[str] = categories if categories else []
        self._category_to_index: Dict[str, int] = {cat: i for i, cat in enumerate(self._categories)}

    def categories(self) -> List[str]:
        """Returns the current list of categories."""
        return self._categories

    def add_category(self, category: str) -> None:
        """Dynamically adds a category if it does not already exist and assigns it an index."""
        if category not in self._category_to_index:
            self._category_to_index[category] = len(self._categories)
            self._categories.append(category)

    def set_categories(self, categories: List[str]) -> None:
        """Sets the categories for the axis and rebuilds the index mapping."""
        self._categories = list(categories)
        self._category_to_index = {cat: idx for idx, cat in enumerate(self._categories)}

    def get_index(self, category: str) -> int:
        """Returns the index of a category, adding it if it doesn't exist."""
        if category not in self._category_to_index:
            self.add_category(category)
        return self._category_to_index[category]

    def get_category(self, index: int) -> str:
        """Returns the category name corresponding to an index."""
        return self._categories[index]

    def to_dict(self) -> Dict:
        """Serializes the axis to a dictionary."""
        return {"name": self.name, "categories": self._categories, "metadata": self.metadata}
    
    @staticmethod
    def from_dict(data: Dict) -> "DustStatAxisConst":
        """Deserializes a DustStatAxisConst from a dictionary."""
        axis = DustStatAxisConst(name=data["name"], categories=data.get("categories", []))
        axis.metadata = data.get("metadata", {})
        axis._category_to_index = {cat: i for i, cat in enumerate(axis._categories)}  # Rebuild index map
        return axis

    def to_yaml(self) -> str:
        """Returns a YAML representation of the axis with category names instead of index keys."""
        formatted_categories = [
            {"cat": cat, "value": self.metadata.get(cat, {}) or {}}
            for cat in self._categories
        ]

        return yaml.dump({
            "name": self.name,
            "categories": formatted_categories,  # Use list format
        }, sort_keys=False)
    
    @staticmethod
    def from_yaml(yaml_str: str) -> "DustStatAxisConst":
        """Deserializes an axis from a YAML string and restores category metadata."""
        data = yaml.safe_load(yaml_str)

        axis = DustStatAxisConst(name=data["name"], categories=[entry["cat"] for entry in data["categories"]])

        # Restore metadata from the new list format
        axis.metadata = {entry["cat"]: entry["value"] for entry in data["categories"] if entry["value"]}

        return axis

class DustStatDataCube(ABC):
    """Abstract base class for a data cube."""

    def __init__(self, name: str, axes: List[DustStatAxisConst], cube_type: DustStatCubeType):
        self.name: str = name
        self._axes: List[DustStatAxisConst] = axes
        self.values: Dict[Tuple[str, ...], Union[int, float, Dict]] = {}
        self.metadata: Dict[str, Union[str, int, float]] = {}
        self.type = cube_type

    def axes(self) -> List[DustStatAxisConst]:
        """Returns the list of axes in the cube."""
        return self._axes

    def set_metadata(self, key: str, value: Union[str, int, float]) -> None:
        """Stores metadata at the cube level."""
        self.metadata[key] = value

    def get_metadata(self, key: str) -> Optional[Union[str, int, float]]:
        """Retrieves metadata at the cube level."""
        return self.metadata.get(key)
    
    def has_value(self, coordinates: Tuple[str, ...]) -> bool:
        """Checks if a value exists at the given coordinates using index-based lookup."""
        try:
            index_tuple = tuple(axis.get_index(cat) for axis, cat in zip(self._axes, coordinates))
            return index_tuple in self.values  # ✅ Use index-based keys
        except ValueError:
            return False  # Category not found in axis, meaning it does not exist in the cube

    def delete_value(self, coordinates: Tuple[str, ...]) -> None:
        """Removes a value from the data cube using category names by converting to index-based keys."""
        index_coords = tuple(axis.get_index(cat) for axis, cat in zip(self._axes, coordinates))
        if index_coords in self.values:
            del self.values[index_coords]

    def to_dict(self, transfer_value_keys = False) -> Dict:
        """Serializes the data cube using index-based storage."""
        if transfer_value_keys:
            return {
                "name": self.name,
                "axes": [axis.to_dict() for axis in self._axes],
                "values": {json.dumps([axis.get_category(idx) for axis, idx in zip(self._axes, key)]): value for key, value in self.values.items()},
                "type": self.type.value,
                "metadata": self.metadata,
            }
        else:
            return {
                "name": self.name,
                "axes": [axis.to_dict() for axis in self._axes],
                "values": {json.dumps(key): value for key, value in self.values.items()},
                "type": self.type.value,
                "metadata": self.metadata,
            }

    def iterate_values(self):
        """Generator that yields (category_names, value) pairs."""
        for coords, value in self.values.items():
            category_names = tuple(axis.get_category(idx) for axis, idx in zip(self._axes, coords))
            yield category_names, value

    def remove_categories(self, categories_per_axis: Dict[DustStatAxis, List[str]]) -> None:
        """
        Removes specified categories from each axis, updates metadata, and reindexes values.

        Args:
            categories_per_axis (Dict[DustStatAxis, List[str]]): 
                A dictionary where keys are axes and values are lists of categories to remove.
        """
        # Step 1: Backup the old category-to-index mapping
        old_category_to_index = {
            axis: {idx: cat for idx, cat in enumerate(axis.categories())} 
            for axis in self._axes
        }

        # Step 2: Remove specified categories from each axis and rebuild mappings
        new_index_mapping_per_axis = {}
        for axis, categories_to_remove in categories_per_axis.items():
            # Keep only categories that are not being removed
            remaining_categories = [cat for cat in axis.categories() if cat not in categories_to_remove]

            # Rebuild category-to-index mapping
            new_index_mapping_per_axis[axis] = {cat: idx for idx, cat in enumerate(remaining_categories)}

            # Remove metadata for deleted categories BEFORE updating categories
            for cat in categories_to_remove:
                if cat in axis.metadata:
                    del axis.metadata[cat]

            # Apply new categories to the axis
            axis.set_categories(remaining_categories)

        # Step 3: Reindex stored values
        new_values = {}
        for old_coords, value in self.values.items():
            new_coords = []
            for idx, axis in enumerate(self._axes):
                old_cat = old_category_to_index[axis][old_coords[idx]]  # Get category name
                if old_cat not in new_index_mapping_per_axis[axis]:  # Skip removed categories
                    break
                new_coords.append(new_index_mapping_per_axis[axis][old_cat])  # Get new index
            else:
                new_values[tuple(new_coords)] = value  # Store only if all categories are valid

        # Step 4: Replace values with updated reindexed data
        self.values = new_values

    def compute_checksum(self, dump_path=None) -> str:
        """Generates a stable SHA-256 checksum by replacing indexes with category names before hashing."""

        def sorted_json(obj):
            """Recursively sorts lists and dictionaries for stable hashing."""
            if isinstance(obj, dict):
                return {k: sorted_json(v) for k, v in sorted(obj.items())}
            elif isinstance(obj, list):
                return sorted((sorted_json(item) for item in obj), key=json.dumps)  # Sort using JSON for stability
            elif isinstance(obj, tuple):
                return tuple(sorted_json(list(obj)))
            else:
                return obj

        # Convert cube to dictionary (uses index-based coords)
        cube_dict = self.to_dict(transfer_value_keys=True)

        # Remove checksum before hashing
        cube_dict["metadata"].pop("checksum", None)

        # Sort everything before hashing
        sorted_cube_dict = sorted_json(cube_dict)

        if dump_path:
            # Write debug file to verify output before hashing
            with open(dump_path, "w", encoding="utf-8") as f:
                json.dump(sorted_cube_dict, f, indent=4, sort_keys=True)

        # Generate checksum
        chechsum = hashlib.sha256(json.dumps(sorted_cube_dict, sort_keys=True).encode()).hexdigest()
        self.set_metadata("checksum", chechsum)

        return chechsum

    @staticmethod
    def from_dict(data: Dict) -> "DustStatDataCube":
        """Deserializes a data cube from a dictionary."""
        axes = [DustStatAxisConst.from_dict(axis_data) for axis_data in data["axes"]]
        cube_type = DustStatCubeType(data["type"])
        
        cube = DustStatDataCubeStructured(data["name"], axes)

        # Convert list keys back to tuples
        cube.values = {tuple(json.loads(key)): value for key, value in data["values"].items()}
        
        return cube
    
    def to_yaml(self) -> str:
        """Returns a YAML representation of the data cube with category names for debugging."""
        formatted_values = [
            {
                "coord": {axis.name: axis.get_category(idx) for axis, idx in zip(self._axes, key)},
                "value": value
            }
            for key, value in self.values.items()
        ]

        formatted_axes = [
            {
                "name": axis.name,
                "categories": [{"cat": category, "value": axis.metadata.get(category, {}) or {}} for category in axis.categories()]
            }
            for axis in self._axes
        ]

        return yaml.safe_dump({
            "name": self.name,
            "axes": formatted_axes,
            "values": formatted_values,
            "type": self.type.value,
            "metadata": self.metadata if self.metadata else {},
        }, sort_keys=False, default_style=None)

    @staticmethod
    def from_yaml(yaml_str: str, safe_load: bool = True) -> "DustStatDataCube":
        """Deserializes a data cube from a YAML string, restoring category names safely."""
        if safe_load:
            data = yaml.safe_load(yaml_str)
        else:
            try:
                from yaml import CLoader as Loader
            except ImportError:
                from yaml import SafeLoader as Loader  # fallback

            data = yaml.load(yaml_str, Loader=Loader)

        # Rebuild axes from YAML (handling {cat, value} format)
        axes = [
            DustStatAxisConst(
                name=axis_data["name"],
                categories=[entry["cat"] for entry in axis_data["categories"]]
            ) for axis_data in data["axes"]
        ]

        # Restore metadata for each axis
        for axis, axis_data in zip(axes, data["axes"]):
            axis.metadata = {entry["cat"]: entry.get("value", {}) for entry in axis_data["categories"] if entry.get("value")}

        # Determine cube type
        cube_type = DustStatCubeType(data["type"])
        cube = DustStatDataCubeStructured(data["name"], axes)

        # Convert category names back to index-based storage
        for entry in data["values"]:
            coord_dict = entry["coord"]
            value = entry["value"]

            # Convert category names to index tuple
            coord_tuple = tuple(axis.get_index(coord_dict[axis.name]) for axis in axes)

            cube.values[coord_tuple] = value  # Restore indexed value

        # ✅ Ensure metadata is restored only when non-empty
        cube.metadata = data.get("metadata", {}) or {}

        return cube


class DustStatDataCubeStructured(DustStatDataCube):
    """Data cube that stores structured (dictionary) values."""

    def __init__(self, name: str, axes: List[DustStatAxisConst]):
        super().__init__(name, axes, DustStatCubeType.STRUCTURED)

    def visit(
        self,
        axis_order: Optional[List[str]] = None,
        sort_order: Optional[Dict[str, SortOrder]] = None,
        filters: Optional[Dict[str, List[str]]] = None
    ) -> Generator[Tuple[Tuple[str, ...], Dict], None, None]:

        axis_name_to_index = {axis.name: i for i, axis in enumerate(self._axes)}

        # Normalize axis order
        if axis_order:
            ordered_axes = [next(axis for axis in self._axes if axis.name == name) for name in axis_order]
            for axis in self._axes:
                if axis not in ordered_axes:
                    ordered_axes.append(axis)
        else:
            ordered_axes = self._axes

        ordered_axis_names = [axis.name for axis in ordered_axes]

        # Build logical coord tuples reordered to axis_order
        coord_map = []
        for index_coords in self.values.keys():
            logical_coords = tuple(axis.categories()[idx] for axis, idx in zip(self._axes, index_coords))

            # Apply axis/category filters
            if filters:
                skip = False
                for axis_name, allowed in filters.items():
                    if allowed is None:
                        continue
                    pos = axis_name_to_index[axis_name]
                    if logical_coords[pos] not in allowed:
                        skip = True
                        break
                if skip:
                    # Skip coord if filter did not allow
                    continue

            reordered_coords = tuple(logical_coords[axis_name_to_index[axis_name]] for axis_name in ordered_axis_names)
            coord_map.append((reordered_coords, logical_coords, index_coords))

        # Sort if sort_order is provided
        if sort_order:
            def coord_sort_key(reordered_coord):
                key = []
                for i, axis_name in enumerate(ordered_axis_names):
                    value = reordered_coord[i]
                    order = sort_order.get(axis_name, SortOrder.ASCENDING)

                    # Reverse value by inverting characters for DESC string sort
                    if order == SortOrder.DESCENDING:
                        value = ''.join(chr(255 - ord(c)) for c in value)

                    key.append(value)
                return key

            coord_map.sort(key=lambda x: coord_sort_key(x[0]))

        # Yield original logical coords and values
        for _, logical_coords, index_coords in coord_map:
            yield logical_coords, self.values[index_coords]

    class Cursor:
        """Cursor for setting and retrieving structured values from a data cube."""

        def __init__(self, cube: "DustStatDataCubeStructured"):
            self.cube = cube
            self.coords: List[Optional[str]] = [None] * len(cube.axes())

        def __get_index_tuple(self) -> Tuple[int, ...]:
            return tuple(
                axis.get_index(self.coords[i]) if self.coords[i] is not None else -1
                for i, axis in enumerate(self.cube.axes())
            )

        def set_coordinate(self, category: str, axis: DustStatAxisConst) -> None:
            """Sets a coordinate value for a given axis, dynamically adding new categories if needed."""
            axis.add_category(category)
            idx_axis = self.cube.axes().index(axis)
            self.coords[idx_axis] = category

        def value(self) -> Optional[Dict]:
            """Retrieves the structured value at the current location."""
            return self.cube.values.get(self.__get_index_tuple(), None)

        def set_value(self, value: Dict) -> None:
            """Sets a structured value at the current location."""
            self.cube.values[self.__get_index_tuple()] = value

        def has_value(self) -> bool:
            """Returns True if a value has been set at the current cursor location (excluding explicit None)."""
            coords_tuple = self.__get_index_tuple()
            return coords_tuple in self.cube.values and self.cube.values[coords_tuple] is not None


import json
import yaml
import deepdiff
from enum import Enum
from datetime import datetime
from dust import Datatypes, ValueTypes, Operation, MetaProps, FieldProps, Committed
from importlib import import_module

import threading

_messages_create_message = None
_store_lock = threading.RLock()

UNIT_ENTITY = "entity"
UNIT_ENTITY_META = "entity_meta"
UNIT_ID = 1
UNIT_META_ID = 2

class UnitMeta(MetaProps):
    name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 100)
    id_cnt = (Datatypes.INT, ValueTypes.SINGLE, 2, 101)
    meta_types = (Datatypes.ENTITY, ValueTypes.SET, 3, 102)

class TypeMeta(MetaProps):
    name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 200)
    fields = (Datatypes.ENTITY, ValueTypes.SET, 2, 201)

class MetaField(MetaProps):
    name = (Datatypes.STRING, ValueTypes.SINGLE, 1, 300)
    global_name = (Datatypes.STRING, ValueTypes.SINGLE, 2, 301)
    field_order = (Datatypes.INT, ValueTypes.SINGLE, 3, 302)
    data_type = (Datatypes.STRING, ValueTypes.SINGLE, 4, 303)
    value_type = (Datatypes.STRING, ValueTypes.SINGLE, 5, 304)

class EntityBaseMeta(MetaProps):
    unit = (Datatypes.ENTITY, ValueTypes.SINGLE, 1, 400)
    meta_type = (Datatypes.ENTITY, ValueTypes.SINGLE, 2, 401)
    entity_id = (Datatypes.INT, ValueTypes.SINGLE, 3, 402)
    committed = (Datatypes.STRING, ValueTypes.SINGLE, 4 ,403)

class EntityTypes(FieldProps):
    type_meta = (UNIT_ENTITY_META, TypeMeta, 1)
    _entity_base = (UNIT_ENTITY_META, EntityBaseMeta, 2)
    unit = (UNIT_ENTITY_META, UnitMeta, 3)
    meta_field = (UNIT_ENTITY_META, MetaField, 4)

class EntityJsonSerializer():
    def dump(self, obj, fp):
        if isinstance(obj, Entity):
            fp.write(json.dumps(obj.to_json()).encode())
            fp.write(b"\n")
        elif isinstance(obj, list):
            objs = []
            for e in obj:
                if isinstance(e, Entity):
                    objs.append(e.to_json())
                else:
                    raise ValueError("Only entity object can be serialized!")
            fp.write(json.dumps(objs).encode())
            fp.write(b"\n")
        else:
            raise ValueError("Only entity object can be serialized!")

    def load(self, f):
        try:
            data = json.loads(f.readline().decode())
            if not isinstance(data, list):
                data = [data]
            return Store.from_json(data)
        except:
            raise ValueError("Invalid json data in messagequeue!")

def get_unit_deps_tuple(module_name, unit_name, meta_type_enums):
    module = import_module(module_name)
    unit_name_attr = getattr(module, unit_name)
    meta_type_enums_attr = getattr(module, meta_type_enums)
    dep_func = getattr(module, "get_unit_dependencies", None)
    return (unit_name_attr, meta_type_enums_attr, dep_func)

def compare_entity_to_json_simple(meta_type_enum, entity, json_entity, json_entity_map, compare_sub_entity_fields_map=None, log_prefix=""):
    changed = {}

    compare_sub_entity_fields = {}
    if compare_sub_entity_fields_map:
        compare_sub_entity_fields = compare_sub_entity_fields_map.get(meta_type_enum, {})

    for field in meta_type_enum.fields_enum:
        if field.valuetype == ValueTypes.SINGLE:
            if field.datatype == Datatypes.ENTITY:
                if compare_sub_entity_fields and field in compare_sub_entity_fields:
                    orig_value = entity.access(Operation.GET, None, field)
                    #print(field.name+"-"+str(orig_value)+"-"+str(_entity_map[entity.global_id()][0]))
                    new_value = None
                    new_entity_global_id = json_entity.get(Store.get_global_fieldname(field))
                    if new_entity_global_id:
                        new_value = json_entity_map[new_entity_global_id]
                        #print(str(new_value))
                    if orig_value is None and not new_value is None or not orig_value is None and new_value is None:
                        changed[log_prefix+field.name] = {"orig_value": orig_value, "new_value": new_value}
                    elif not orig_value is None and not new_value is None:
                        changed.update(compare_entity_to_json_simple(compare_sub_entity_fields[field], orig_value, new_value, json_entity_map, compare_sub_entity_fields_map, log_prefix="{}#".format(field.name)))
            else:
                entity_changed = False
                if ( json_entity is None and not entity is None ) or ( not json_entity is None and entity is None ):
                    entity_changed = True
                elif json_entity and entity:
                    value1 = entity.access(Operation.GET, None, field)
                    value2 = json_entity.get(Store.get_global_fieldname(field))
                    if value1 is None and not value2 is None or not value1 is None and value2 is None:
                        entity_changed = True
                    elif value1 and value2:
                        if field.datatype == Datatypes.BOOL:
                            entity_changed = bool(value1) != bool(value2)
                        elif field.datatype == Datatypes.INT:
                            entity_changed = int(value1) != int(value2)
                        elif field.datatype == Datatypes.NUMERIC:
                            entity_changed = abs(float(value1) - float(value2)) > 0.0001
                        elif field.datatype == Datatypes.STRING:
                            entity_changed = str(value1) != str(value2)
                        elif field.datatype == Datatypes.JSON:
                            dd = deepdiff.DeepDiff(value1, value2, ignore_order=True)
                            if dd:
                                entity_changed = True
                        else:
                            entity_changed = value1 != value2
                    
                if entity_changed:
                    changed[log_prefix+field.name] = {"orig_value": entity.access(Operation.GET, None, field), "new_value": json_entity.get(Store.get_global_fieldname(field))}

        elif field.valuetype == ValueTypes.SET or field.valuetype == ValueTypes.LIST or field.valuetype == ValueTypes.MAP:
            if field.datatype != Datatypes.ENTITY:
                iter1 = entity.access(Operation.GET, None, field)
                iter2 = json_entity.get(Store.get_global_fieldname(field))
                if field.valuetype == field.valuetype == ValueTypes.SET and iter2:
                    iter2 = set(iter2)

                dd = deepdiff.DeepDiff(iter1, iter2, ignore_order=True)
                if dd:
                    orig_value = entity.access(Operation.GET, None, field)
                    new_value = json_entity.get(Store.get_global_fieldname(field))

                    if field.valuetype == ValueTypes.SET:
                        if orig_value:
                            orig_value = list(orig_value)
                        if new_value:
                            new_value = list(new_value)
                    changed[log_prefix+field.name] = {"orig_value": orig_value, "new_value": new_value}
            elif field.valuetype == ValueTypes.SET or field.valuetype == ValueTypes.LIST:
                if compare_sub_entity_fields and field in compare_sub_entity_fields:
                    orig_value_global_ids = entity.access(Operation.GET, None, field)
                    new_entity_global_ids = json_entity.get(Store.get_global_fieldname(field))
                    # at this point we only have unordered ids, so we have to compare everything to everything and fail as early as possible
                    if not orig_value_global_ids and new_entity_global_ids or not new_entity_global_ids and orig_value_global_ids or len(orig_value_global_ids) != len(new_entity_global_ids):
                        changed[log_prefix+field.name] = {"orig_list_value": list(orig_value_global_ids), "new_list_value": list(new_entity_global_ids)}
                    else:
                        # Same number of entries
                        no_match_found = False
                        match_to = list(new_entity_global_ids)
                        for orig_value_global_id in orig_value_global_ids:
                            orig_value = Store.access(Operation.GET, None, orig_value_global_id)
                            if orig_value is None:
                                no_match_found = True
                                break
                            else:
                                match_found = False
                                for new_entity_global_id in match_to:
                                    new_value = json_entity_map[new_entity_global_id]
                                    if not new_value is None:
                                        sub_changed = compare_entity_to_json_simple(compare_sub_entity_fields[field], orig_value, new_value, json_entity_map, compare_sub_entity_fields_map, log_prefix="{}#".format(field.name))
                                        if len(sub_changed) == 0:
                                            match_found = True
                                            break
                                if not match_found:
                                    no_match_found = True
                                    break
                        if no_match_found:
                            changed[log_prefix+field.name] = {"orig_list_value": list(orig_value_global_ids), "new_list_value": list(new_entity_global_ids)}

    return changed

class Store():
    @staticmethod
    def load_unit_types(filenames):
        for filename in filenames:
            with open(filename, "r") as tf:
                loaded__meta_types = yaml.load(tf, Loader=yaml.FullLoader)["types"]

            Store.load_types_from_dict(loaded__meta_types)

    @staticmethod
    def create_unit(unit_name, unit_id):
        entity_map = globals()["_entity_map"]
        enum_map = globals()["_enum_map"]

        if not unit_name in [UNIT_ENTITY, UNIT_ENTITY_META] and not unit_name in enum_map:
            unit = Store.access(Operation.GET, None, UNIT_ENTITY, unit_id, EntityTypes.unit)
            enum_map[unit_name] = unit
            enum_map[unit] = unit_name
            #print("Unit name: {} ({}), global id: {}".format(unit_name, unit.entity_id, unit.global_id()))
            unit.access(Operation.SET, 0, UnitMeta.id_cnt)
            unit.access(Operation.SET, unit_name, UnitMeta.name)
        
        elif not ( UNIT_ENTITY+":2:"+EntityTypes.unit.name ) in entity_map:
            unit = Store._create_entity(None, 1, None)
            unit.unit = unit
            enum_map[UNIT_ENTITY] = unit
            enum_map[unit] = UNIT_ENTITY

            unit_meta = Store._create_entity(unit, 2, None)
            enum_map[UNIT_ENTITY_META] = unit_meta
            enum_map[unit_meta] = UNIT_ENTITY_META

            cnt = 0
            types = {}
            for entity_type in EntityTypes:
                cnt += 1
                et = Store._create_entity(unit_meta, entity_type.id_value, None)
                enum_map[entity_type] = et
                enum_map[et] = entity_type
                if entity_type.id_value > cnt:
                    cnt = entity_type.id_value

            unit.meta_type = enum_map[EntityTypes.unit]
            unit_meta.meta_type = enum_map[EntityTypes.unit]
            enum_map[EntityTypes.unit].meta_type = enum_map[EntityTypes.type_meta]

            Store._add_entity_to_store(unit)
            Store._add_entity_to_store(unit_meta)

            for entity_type in [EntityTypes.type_meta, EntityTypes.unit, EntityTypes._entity_base, EntityTypes.meta_field]:
                et = enum_map[entity_type]
                et.meta_type = enum_map[EntityTypes.type_meta]
                enum_map[et.global_id()] = entity_type
                Store._add_entity_to_store(et)
                et.access(Operation.SET, entity_type.name, TypeMeta.name)

            unit.access(Operation.SET, 2, UnitMeta.id_cnt)
            unit.access(Operation.SET, UNIT_ENTITY, UnitMeta.name)

            unit_meta.access(Operation.SET, cnt, UnitMeta.id_cnt)
            unit_meta.access(Operation.SET, UNIT_ENTITY_META, UnitMeta.name)

            return unit_meta

        else:
            unit = enum_map[unit_name]

        return unit

    @staticmethod
    def _get_base_fields():
        fields = []
        for base_field in EntityBaseMeta:
            fields.append(Store._global_field_name(UNIT_ENTITY_META, EntityTypes._entity_base.name, base_field.name))

        return tuple(fields)

    @staticmethod
    def _create_entity(unit, entity_id, meta_type):
        enum_map = globals()["_enum_map"]

        if unit and not isinstance(unit, Entity):
            unit = enum_map[unit]
        if meta_type and not isinstance(meta_type, Entity):
            meta_type = enum_map[meta_type]

        e = Entity(unit, entity_id, meta_type)

        if unit and meta_type:
            Store._add_entity_to_store(e)

        return e

    @staticmethod
    def _add_entity_to_store(e):
        entity_map = globals()["_entity_map"]

        unit_field, meta_type_field, entity_id_field, committed_field = Store._get_base_fields()

        entity_map[e.global_id()] = ({
            unit_field: e.unit.global_id(), 
            entity_id_field: e.entity_id, 
            meta_type_field: e.meta_type.global_id(),
            committed_field: e.committed.name
        }, e)

    @staticmethod
    def _delete_entity_to_store(e):
        entity_map = globals()["_entity_map"]

        if e.global_id() in entity_map:
            del entity_map[e.global_id()]

    @staticmethod
    def increment_unit_counter(unit, requested_id):
        if isinstance(unit, str):
            unit = globals()["_enum_map"][unit]
        unit_entity_map = globals()["_entity_map"][unit.global_id()][0]

        if requested_id is None:
            return unit.access(Operation.CHANGE, 1, UnitMeta.id_cnt)
        elif not unit.entity_id is None:
            current_cnt = unit_entity_map.get("entity_meta:unit:id_cnt")
            if current_cnt == None or current_cnt < requested_id:
                unit_entity_map["entity_meta:unit:id_cnt"] = requested_id

            return requested_id

    @staticmethod
    def load_types_from_enum(e, unit_meta_id):
        enum_map = globals()["_enum_map"]
        unit_meta_types = {}
        for meta_type in e:
            for field in meta_type.fields_enum:
                global_name = Store._global_field_name(meta_type.unit_name, meta_type.name, field.name)
                enum_map[global_name] = {"datatype": field.value[0], "valuetype": field.value[1], "id": field.id_value, "field_order": field.order_value, "_enum": field}
                enum_map[field] = global_name

        for base_field in EntityBaseMeta:
            global_name = Store._global_field_name(UNIT_ENTITY_META, EntityTypes._entity_base.name, base_field.name)
            enum_map[global_name] = {"datatype": base_field.value[0], "valuetype": base_field.value[1], "id": base_field.id_value, "field_order": base_field.order_value, "_enum": base_field}
            enum_map[base_field] = global_name

        for meta_type in e:
            unit = Store.create_unit(meta_type.unit_name, unit_meta_id)

            if meta_type in enum_map:
                field_meta_type = enum_map[meta_type]
            else:
                cnt = unit.access(Operation.GET, 0, UnitMeta.id_cnt)
                if meta_type.id_value > cnt:
                    unit.access(Operation.SET, meta_type.id_value, UnitMeta.id_cnt)
                field_meta_type = Store.access(Operation.GET, None, meta_type.unit_name, meta_type.id_value, EntityTypes.type_meta)
                field_meta_type.access(Operation.SET, meta_type.name, TypeMeta.name)
                enum_map[meta_type] = field_meta_type
                enum_map[field_meta_type] = meta_type
                enum_map[field_meta_type.global_id()] = meta_type

            field_entities = []
            max_id_value = 0
            for field in meta_type.fields_enum:
                global_name = Store._global_field_name(meta_type.unit_name, meta_type.name, field.name)

                field_entity = Store.access(Operation.GET, None, meta_type.unit_name, field.id_value, EntityTypes.meta_field)
                field_entity.access(Operation.SET, field.name, MetaField.name)
                field_entity.access(Operation.SET, field.order_value, MetaField.field_order)
                field_entity.access(Operation.SET, global_name, MetaField.global_name)
                field_entity.access(Operation.SET, field.datatype.name, MetaField.data_type)
                field_entity.access(Operation.SET, field.valuetype.name, MetaField.value_type)
                field_entities.append(field_entity)

                if field.id_value > max_id_value:
                    max_id_value = field.id_value

            for field_entity in field_entities:
                field_meta_type.access(Operation.ADD, field_entity, TypeMeta.fields)

            if max_id_value > 0:
                cnt = unit.access(Operation.GET, 0, UnitMeta.id_cnt)
                if max_id_value > cnt:
                    unit.access(Operation.SET, max_id_value, UnitMeta.id_cnt)

            unit_meta_types.setdefault(unit, []).append(field_meta_type)

        for unit, field_meta_types in unit_meta_types.items():
            for field_meta_type in field_meta_types:
                unit.access(Operation.ADD, field_meta_type, UnitMeta.meta_types)

    @staticmethod
    def access(operation, value, *path):
        with _store_lock:
            return Store._access(operation, value, *path)

    @staticmethod
    def _access(operation, value, *path):
        entities = globals()["_entity_map"]
        enum_map = globals()["_enum_map"]

        local_ref = None
        last_obj = None
        remaining_path = None
        idx = 0
        last_global_id = None
        last_entity_path = []

        unit = None
        meta_type = None

        if len(path) > 0:
            if operation == Operation.WALK and isinstance(path[0], list):
                last_obj = path[0]
                idx = 1
            else:
                unit, entity_id, meta_type = Entity._resolve_global_id(path[0])
                if entity_id is None:
                    # path is done with 3 parts
                    if len(path) > 2 and ( path[1] is None or isinstance(path[1], int) ) and \
                        ( isinstance( path[2], FieldProps ) or isinstance( path[2], Entity ) ):
                        entity_id = Store.increment_unit_counter(path[0], path[1])
                        unit = path[0]
                        meta_type = path[2]
                        local_ref = Entity._ref(unit, entity_id, meta_type)
                        idx = 3
                else:
                    # found entity id on path[0]
                    local_ref = path[0]
                    idx = 1

        if idx < len(path):
            remaining_path = path[idx:]
        else:
            remaining_path = []

        #print("Access 1: local_ref={}, path={}".format(local_ref, path))
        if local_ref:
            if not local_ref in entities:
                last_obj = Store._create_entity(unit, entity_id, meta_type)
            else:
                last_obj = entities[local_ref][1]
            last_global_id = last_obj.global_id()

        path_length = len(remaining_path)
        if last_obj and path_length == 0 and operation == Operation.GET:
            if _messages_create_message:
                _create_message(MessageType.ENTITY_ACCESS, {"path": last_entity_path, "op": operation}, [last_global_id])
            return last_obj
        else:
            rv = None

            if last_obj is None:
                last_obj = [e[1] for e in entities.values()]
                parent = last_obj

            for last_idx in range(path_length):
                key = remaining_path[last_idx]
                #print("{}: Access 2: key={}".format(last_obj, key))

                if last_obj is None:
                    last_obj = Store._access_data_create_container(parent, remaining_path[last_idx - 1], key)
                elif isinstance(last_obj, Entity):
                    last_global_id = last_obj.global_id()
                    last_entity_path = path[last_idx:]

                if isinstance(last_obj, str):
                    unit, entity_id, meta_type = Entity._resolve_global_id(last_obj)
                    if entity_id:
                        parent = entities[last_obj][1]
                    else:
                        parent = last_obj
                else:
                    parent = last_obj

                last_obj = Store._access_data_get_value(parent, key, None)
                last_idx += 1

            if operation == Operation.SET or operation == Operation.ADD or operation == Operation.CHANGE:
                changed, rv = Store._access_data_set_value(parent, remaining_path[last_idx - 1], value, operation)
                if changed and _messages_create_message:
                    _create_message(MessageType.ENTITY_ACCESS, {"path": last_entity_path, "op": operation}, [last_global_id])

            elif operation == Operation.GET:
                #if _messages_create_message:
                #    _create_message(MessageType.ENTITY_ACCESS, {"path": last_entity_path, "op": operation}, [last_global_id])
                rv = last_obj

            elif operation == Operation.VISIT:
                #if _messages_create_message:
                #    _create_message(MessageType.ENTITY_ACCESS, {"path": last_entity_path, "op": operation}, [last_global_id])
                rv = last_obj

                if callable(value):
                    if isinstance(last_obj, list) or isinstance(last_obj, set):
                        for e in last_obj:
                            value(e)
                    elif isinstance(last_obj, dict):
                        for key, value in last_obj.items():
                            value(key, value)

            elif operation == Operation.WALK:
                stored_entities = set()
                if isinstance(last_obj, list):
                    rv = last_obj
                else:
                    rv = [last_obj]
                Store._walk_entities(rv, stored_entities)

            return rv

    @staticmethod
    def _walk_entities(entities, stored_entities):
        idx = 0
        while True:
            if idx >= len(entities):
                break

            entity = entities[idx]
            idx += 1
                
            meta_type = Store.get_meta_type_enum(entity)
            field_enums = [EntityTypes._entity_base.fields_enum, meta_type.fields_enum]
            for field_enum in field_enums:
                for field in field_enum:
                    if field.datatype == Datatypes.ENTITY:
                        if field.valuetype == ValueTypes.SINGLE:
                            value = entity.access(Operation.GET, None, field)
                            if isinstance(value, Entity) and not value.global_id() in stored_entities:
                                entities.append(value)
                                stored_entities.add(value.global_id())
                        elif field.valuetype in [ValueTypes.LIST, ValueTypes.SET]:
                            values = entity.access(Operation.GET, None, field)
                            if values:
                                for value in values:
                                    if value and not value in stored_entities:
                                        list_entity = Store.access(Operation.GET, None, value)
                                        if list_entity is None:
                                            print(entity.global_id() + " - " + value)
                                        entities.append(list_entity)
                                        stored_entities.add(list_entity.global_id())

    @staticmethod
    def _access_data_create_container(obj, field, key):
        if isinstance(obj, Entity):
            entities = globals()["_entity_map"]
            enum_map = globals()["_enum_map"]

            global_field_name = enum_map[field]
            field_config = enum_map[global_field_name]

            e_map = entities[obj.global_id()][0]
            if field_config["valuetype"] == ValueTypes.SET:
                e_map[global_field_name] = set()
            elif field_config["valuetype"] == ValueTypes.LIST:
                e_map[global_field_name] = []
            
            return e_map[global_field_name]

    @staticmethod
    def _set_uncommitted(e, e_map):
        if e.committed == Committed.SAVED:
            e.committed = Committed.UPDATED

        committed_field = Store._get_base_fields()[3]
        e_map[committed_field] = e.committed.name

        return True

    @staticmethod
    def _access_data_set_value(obj, field, value, operation):
        if isinstance(obj, Entity):
            changed = False

            entities = globals()["_entity_map"]
            enum_map = globals()["_enum_map"]

            e_map = entities[obj.global_id()][0]
            entity = entities[obj.global_id()][1]
            global_field_name = enum_map[field]
            field_config = enum_map[global_field_name]

            if field.datatype == Datatypes.ENTITY and isinstance(value, Entity):
                if field.valuetype == ValueTypes.LIST:
                    changed = Store._set_uncommitted(entity, e_map)
                    e_map.setdefault(global_field_name, []).append(value.global_id())
                elif field.valuetype == ValueTypes.SET:
                    if not global_field_name in e_map or not value.global_id in e_map[global_field_name]:
                        e_map.setdefault(global_field_name, set()).add(value.global_id())
                        changed = Store._set_uncommitted(entity, e_map)
                else:
                    if e_map.get(global_field_name, None) != value: # TODO: change if it shoudl be value.global_id()
                        e_map[global_field_name] = value.global_id()
                        changed = Store._set_uncommitted(entity, e_map)
            else:
                if field.valuetype == ValueTypes.LIST and not isinstance(value, list):
                    changed = Store._set_uncommitted(entity, e_map)
                    e_map.setdefault(global_field_name, []).append(value)
                elif field.valuetype == ValueTypes.SET and not isinstance(value, set):
                    if isinstance(value, list):
                        all_included = False
                        if global_field_name in e_map:
                            all_included = True
                            old_values = e_map[global_field_name]
                            for v in value:
                                all_included = v in old_values
                                if not all_included:
                                    break
                        if not all_included:
                            e_map.setdefault(global_field_name, set()).update(value)
                            changed = Store._set_uncommitted(entity, e_map)
                    else:
                        if not global_field_name in e_map or not value in e_map[global_field_name]:
                            e_map.setdefault(global_field_name, set()).add(value)
                            changed = Store._set_uncommitted(entity, e_map)
                else:
                    if operation == Operation.CHANGE:
                        e_map[global_field_name] += value
                        value = e_map[global_field_name]
                        changed = Store._set_uncommitted(entity, e_map)
                    else:
                        if e_map.get(global_field_name, None) != value:
                            e_map[global_field_name] = value
                            changed = Store._set_uncommitted(entity, e_map)

            return (changed, value)

        return (False, None)

    @staticmethod
    def _access_data_get_value(obj, key, default_value):
        entities = globals()["_entity_map"]
        enum_map = globals()["_enum_map"]

        if isinstance(obj, Entity):
            global_name = enum_map[key]
            e_map = entities[obj.global_id()][0]
            if global_name in e_map:
                unit, entity_id, meta_type = Entity._resolve_global_id(e_map[global_name])
                if entity_id:
                    if not e_map[global_name] in entities:
                        e = Store.access(Operation.GET, None, e_map[global_name])
                        #print(str(e.global_id()))
                    return entities[e_map[global_name]][1]
                else:
                    return e_map[global_name]
        elif isinstance(obj, list):
            if isinstance(key, int) and key < len(obj):
                return obj[key]

        elif isinstance(obj, dict):
            if key in obj:
                return obj[key]

        elif isinstance(obj, set):
            raise Exception("Not possible to request path element on sets.")

        return default_value

    @staticmethod
    def _get_field_config(unit_name, meta_type, field):
        enum_map = globals()["_enum_map"]
        if isinstance(meta_type, Enum) and meta_type in [EntityTypes.type_meta, EntityTypes.meta_field]:
            return enum_map[UNIT_ENTITY_META+":"+meta_type.name+":"+field.name]
        elif isinstance(meta_type, str) and meta_type in [EntityTypes.type_meta.name, EntityTypes.meta_field.name]:
            return enum_map[UNIT_ENTITY_META+":"+meta_type+":"+field]
        elif isinstance(meta_type, Enum) and isinstance(field, Enum):
            return enum_map[unit_name+":"+meta_type.name+":"+field.name]
        else:
            return enum_map[unit_name+":"+meta_type+":"+field]

    @staticmethod
    def _global_field_name(unit, meta_type, field_name):
        if isinstance(unit, Entity) and isinstance(meta_type, Entity):
            return  enum_map[unit]+":"+enum_map[meta_type].name+":"+field_name
        else:
            return "{}:{}:{}".format(unit, meta_type, field_name)

    @staticmethod
    def to_json(*path):
        entities = Store.access(Operation.VISIT, None, *path)
        json_array = []
        for e in entities:
            json_array.append(e.to_json())
        return json_array

    @staticmethod
    def from_json(objects):
        entities = []
        for e_map in objects:
            entities.append(Entity.from_json(e_map))

        return entities

    @staticmethod
    def get_meta_type_enum(entity):
        enum_map = globals()["_enum_map"]
        if isinstance(entity, Entity):
            return enum_map[entity.meta_type]
        elif isinstance(entity, dict):
            _, meta_type_field, _,  _ = Store._get_base_fields()
            return enum_map[entity[meta_type_field]]

    @staticmethod
    def get_global_fieldname(field):
        enum_map = globals()["_enum_map"]
        return enum_map[field]

    @staticmethod
    def global_id(entity):
        if isinstance(entity, Entity):
            return entity.global_id()
        elif isinstance(entity, dict):
            unit_field, meta_type_field, entity_id_field,  _ = Store._get_base_fields()
            unit = Store.access(Operation.GET, None, entity[unit_field])
            meta_type = Store.access(Operation.GET, None, entity[meta_type_field])
            return Entity._ref(unit, entity[entity_id_field], meta_type)

    @staticmethod
    def concat_field_values(entity, fields, separator="#", ext_store=None):
        values = []
        template = separator.join(["{}" for i in range(len(fields))])
        for field in fields:
            if isinstance(entity, Entity):
                if isinstance(field, Enum):
                    values.append(entity.access(Operation.GET, None, field))
                elif isinstance(field, tuple):
                    values.append(entity.access(Operation.GET, None, *field))
            elif isinstance(entity, dict):
                if isinstance(field, Enum):
                    values.append(entity.get(Store.get_global_fieldname(field)))
                elif isinstance(field, tuple):
                    # Only works if except last field all are Entities
                    e = entity
                    v = None
                    for field_element in field:
                        if e and field_element.datatype == Datatypes.ENTITY:
                            e = ext_store.get(e.get(Store.get_global_fieldname(field_element)))
                        else:
                            v = e.get(Store.get_global_fieldname(field_element))
                    values.append(v)

        return template.format(*values)

class Entity():
    def __init__(self, unit, entity_id, meta_type):
        self.unit = unit
        self.entity_id = entity_id
        self.meta_type = meta_type
        self.committed = Committed.CREATED

    def access(self, operation, value, *path):
        enum_map = globals()["_enum_map"]
        return Store.access(operation, value, enum_map[self.unit], self.entity_id, enum_map[self.meta_type], *path)

    @staticmethod
    def _ref(unit, entity_id, meta_type):
        enum_map = globals()["_enum_map"]
        if isinstance(unit, Entity) and isinstance(meta_type, Entity):
            return  enum_map[unit]+":"+str(entity_id)+":"+enum_map[meta_type].name
        else:
            return  unit+":"+str(entity_id)+":"+meta_type.name

    def global_id(self):
        return  Entity._ref(self.unit, self.entity_id, self.meta_type)

    def _resolve_global_id(value):
        if isinstance(value, str):
            enum_map = globals()["_enum_map"]
            parts = value.split(":")
            entity_id = None
            try:
                entity_id = int(parts[1])
            except:
                pass
            if len(parts) == 3 and not entity_id is None:
                entity = globals()["_entity_map"].get(value, (None, None))[1]
                unit = enum_map.get(parts[0], None)
                if unit and isinstance(unit, Entity) and enum_map[unit.meta_type] == EntityTypes.unit:
                    if entity and entity.meta_type and isinstance(entity.meta_type, Entity) and entity.meta_type in enum_map:
                        return (unit, entity_id, entity.meta_type)
                    elif entity is None:
                        unit_meta = enum_map[unit.access(Operation.GET, None, UnitMeta.name)+"_meta"]
                        for meta_type_global_id in unit_meta.access(Operation.GET, None, UnitMeta.meta_types):
                            meta_type_entity = Store.access(Operation.GET, None, meta_type_global_id)
                            if meta_type_entity.access(Operation.GET, None, TypeMeta.name) == parts[2]:
                                #print("None existant entity, but create it for {}. Type: {}".format(value, meta_type_entity.access(Operation.GET, None, TypeMeta.name)))
                                return (unit, entity_id, meta_type_entity)


        return (None, None, None)

    def get_meta_type_enum(self):
        enum_map = globals()["_enum_map"]
        return _enum_map[self.meta_type]

    def __to_json(self, json_map):
        entity_value_map = globals()["_entity_map"][self.global_id()][0]
        enum_map = globals()["_enum_map"]

        for field, value in entity_value_map.items():
            parts = field.split(":")
            field_config = Store._get_field_config(parts[0], parts[1], parts[2])

            if field_config["valuetype"] in [ValueTypes.SET, ValueTypes.LIST] and ( isinstance(value, list) or isinstance(value, set) ):
                json_map[field] = []
                for element in value:
                    json_map[field].append(element)
            else:
                json_map[field] = value

    def to_json(self):
        json_map = {}
        self.__to_json(json_map)
        return json_map

    def set_committed(self):
        entity_map = globals()["_entity_map"]
        self.committed = Committed.SAVED

        _, _, _, committed_field = Store._get_base_fields()

        entity_map[self.global_id()][0][committed_field] = self.committed.name

    @staticmethod
    def from_json(e_map, force_new_id=False, force_entity_id=None):
        entity_map = globals()["_entity_map"]
        enum_map = globals()["_enum_map"]

        unit_field, meta_type_field, entity_id_field, committed_field = Store._get_base_fields()

        unit_parts = e_map[unit_field].split(":")

        unit = Store.access(Operation.GET, None, UNIT_ENTITY, int(unit_parts[1]), EntityTypes.unit)
        unit_name = unit.access(Operation.GET, None, UnitMeta.name)
        meta_type = enum_map[e_map[meta_type_field]]

        e = None
        if not force_entity_id is None:
            e = Store.access(Operation.GET, None, force_entity_id)

        if e is None:
            if force_new_id:
                entity_id = None
            else:
                entity_id = int(e_map[entity_id_field])

            e = Store.access(Operation.GET, None, unit_name, entity_id, meta_type)

        # Wipe none base fields
        orig_map = entity_map[e.global_id()][0]
        del_fields = []
        for field in orig_map.keys():
            if not field in [unit_field, meta_type_field, entity_id_field, committed_field]:
                del_fields.append(field)
        for field in del_fields:
            del orig_map[field]

        for field, value in e_map.items():
            if not field in [unit_field, meta_type_field, entity_id_field, committed_field]:
                field_parts = field.split(":")
                field_config = Store._get_field_config(field_parts[0], field_parts[1], field_parts[2])
                if field_config["valuetype"] == ValueTypes.SET and isinstance(value, list):
                    e.access(Operation.SET, set(value), field_config["_enum"])
                else:
                    e.access(Operation.SET, value, field_config["_enum"])
            elif field == committed_field:
                if Committed[value] == Committed.SAVED:
                    e.set_committed()
        return e

    def delete(self):
        Store._delete_entity_to_store(self)

_entity_map = {}
_enum_map = {}

Store.load_types_from_enum(EntityTypes, UNIT_META_ID)
_messages_module = import_module("dust.messages")
MessageType = getattr(_messages_module, "MessageType")
MessageTypes = getattr(_messages_module, "MessageTypes")
_messages_create_message = getattr(_messages_module, "create_message")
_UNIT_MESSAGE = _messages_module.UNIT_MESSAGES

def _create_message(message_type, message_params, entities):
    if _messages_create_message and message_params["op"] in [Operation.ADD, Operation.SET, Operation.DEL, Operation.CHANGE]:
        params = {}
        if entities and not entities[0] is None:
            enum_map = globals()["_enum_map"]
            unit, _, _ = Entity._resolve_global_id(entities[0])
            unit_name = enum_map[unit]
            if unit_name in [_UNIT_MESSAGE, UNIT_ENTITY] or unit_name.endswith("_meta"):
                return

        params["op"] = message_params["op"].name
        params["path"] = []
        for path_element in message_params["path"]:
            if isinstance(path_element, Enum):
                params["path"].append(path_element.name)
            elif isinstance(path_element, Entity):
                params["path"].append(path_element.global_id())

        _messages_create_message(message_type, params, entities)

if __name__ == '__main__':
    print(json.dumps(_entity_map, indent=4))
    print(str(_meta_ref))



from jinja2 import Template
from enum import Enum
import traceback
import json
import time
import io
import datetime
import os

from dust import Datatypes, ValueTypes, Operation, Committed
from dust.entity import UNIT_ENTITY, EntityTypes, EntityBaseMeta, Store, UnitMeta
from dust.events import FORMAT_DB_DATETIME

_sql_persister = None
_types_initiated = set()

class TrxState:
    execute = 0
    commitstarted = 1
    commitfinished = 2
    rollbackstarted = 3
    rollbackfinished = 4

DB_UPDATE_BATCH = int(os.environ.get("DB_UPDATE_BATCH", 50000))
DELAY_IN_SEC_AFTER_UPDATE = float(os.environ.get("DELAY_IN_SEC_AFTER_UPDATE", 0.0))
DEFAULT_MAX_PACKET_SIZE = 4 * 1024 * 1024

def init_sql_persist(unit_name, persist_class, meta_type_enums, deps_func):
    global _sql_persister
    global _types_initiated

    if not meta_type_enums in _types_initiated:
        print("Persist: Initiating {}/{}".format(unit_name, meta_type_enums.__name__))

        if _sql_persister is None:
            _sql_persister = persist_class()

        schemas = _sql_persister.generate_schema(unit_name, meta_type_enums)
        if schemas:
            print(f"{unit_name} meta has changed")
            for e in Store.access(Operation.VISIT, None):
                if e.get_meta_type_enum() == EntityTypes.unit and e.access(Operation.GET, None, UnitMeta.name) == f"{unit_name}_meta":
                    unit_meta_entity = e;
                    meta_persist_entities = [pe for pe in Store.access(Operation.WALK, None, [e]) if pe.unit == unit_meta_entity]
                    meta_persist_entities.append(unit_meta_entity)

                    # delete schema for unit_meta
                    delete_map = {}
                    # Delete entity_unit
                    print(f"Deleting meta in unit {e.global_id()}/{unit_name}")
                    _sql_persister._SqlPersist__prepare_delete_updating_meta_records(delete_map, "entity_unit", [("_global_id", "=", unit_meta_entity.global_id())])
                    _sql_persister._SqlPersist__prepare_delete_updating_meta_records(delete_map, "entity_unit_meta_types", [("_global_id", "=", unit_meta_entity.global_id())])
                    _sql_persister._SqlPersist__prepare_delete_updating_meta_records(delete_map, "entity_type_meta", [("_unit", "=", unit_meta_entity.global_id())])
                    _sql_persister._SqlPersist__prepare_delete_updating_meta_records(delete_map, "entity_type_meta_fields", [("_global_id", "like", f"{unit_name}_meta:%:type_meta")])
                    _sql_persister._SqlPersist__prepare_delete_updating_meta_records(delete_map, "entity_meta_field", [("_global_id", "like", f"{unit_name}_meta:%:meta_field")])
                    _sql_persister._SqlPersist__delete_updating_meta_records(delete_map)

                    persist_entities(meta_persist_entities)

        _types_initiated.add(meta_type_enums)

        if deps_func:
            unit_dependencies = deps_func()
            if unit_dependencies:
                for dep_unit_name, dep_meta_type_enums, dep_deps_func in unit_dependencies:
                    init_sql_persist(dep_unit_name, persist_class, dep_meta_type_enums, dep_deps_func)

    _sql_persister.print_table_alterations(unit_name)

def load_all():
    global _sql_persister
    return _sql_persister.load_all()    

def load_units():
    global _sql_persister
    return _sql_persister.load_units()    

def load_unit_type(unit_meta_type, where_filters=None, load_referenced=False, entity_filter_method=None):
    global _sql_persister
    return _sql_persister.load_type(unit_meta_type, where_filters=where_filters, load_referenced=load_referenced, entity_filter_method=entity_filter_method)    

def load_entity_ids_for_type(unit_meta_type, where_filters=None):
    global _sql_persister
    return _sql_persister.load_entity_ids_for_type(unit_meta_type, where_filters=where_filters)    

def persist_entities(entities, trx_state_callback=None):
    global _sql_persister
    _sql_persister.persist_entities(entities, trx_state_callback)   

def  dump_database(stream):
    global _sql_persister
    _sql_persister.dump_database(stream)

BASE_FIELD_NAMES = [f"_{base_field.name}" for base_field in EntityTypes._entity_base.fields_enum if base_field != EntityBaseMeta.committed]
BASE_FIELD_NAMES.insert(0, "_global_id")
MULTIVALUE_BASE_FIELD_NAMES = ["_global_id", "_value_cnt"]

ALL_TABLES = {}

class SqlField():
    def __init__(self, field_name, field_type, primary_key=False, base_field=False):
        self.field_name = field_name
        self.field_type = field_type
        self.primary_key = primary_key
        self.base_field = base_field

class SqlTable():
    
    def __init__(self, table_name):
        self.table_name = table_name
        self.fields = []
        self.primary_keys = []
        self.value_tables_by_fields = {}
        self.multi_value_table = False

    def add_field(self, sql_field, sql_type, primary_key=False, base_field=False):
        field = SqlField(sql_field, sql_type, primary_key, base_field)
        self.fields.append(field)
        if primary_key:
            self.primary_keys.append(field)

class SqlPersist():
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.__persisted_types = set()
        self.__unit_meta_unit = {}

    def table_exits(self, table_name, conn):
        pass 

    def sql_type(self, datatype, valuetype, primary_key=False):
        pass

    def create_table_template(self, sql_table):
        pass 

    def alter_table_template(self):
        pass 

    def delete_table_template(self):
        pass 

    def alter_table(self, sql, conn):
        pass

    def insert_into_table_template(self):
        pass

    def select_template(self, where_filters):
        pass

    def update_template(self):
        pass

    def delete_template(self, where_filters=None):
        pass

    def convert_value_to_db(self, field, value):
        pass

    def create_cursor(self, conn):
        pass

    def close_cursor(self, conn):
        pass

    def create_exectute_params(self):
        pass

    def add_execute_param(self, values, name, value, operator="=", named_param=True):
        pass

    def map_value_to_db(self, field, entity):
        value = entity.access(Operation.GET, None, field)
        if value is None:
            return None 
        else:
            if field.valuetype == ValueTypes.SINGLE:
                return self.convert_value_to_db(field, value)

            elif field.valuetype == ValueTypes.SET:
                return [self.convert_value_to_db(field, v) for v in value]

            elif field.valuetype == ValueTypes.LIST and field.datatype != Datatypes.JSON:
                return [self.convert_value_to_db(field, v) for v in value]

            elif field.valuetype == ValueTypes.MAP or field.valuetype == ValueTypes.LIST:
                return json.dumps(value)

    def map_value_from_db(self, field, value):
        if value is None:
            return None

        else:
            if field.valuetype in [ValueTypes.SINGLE, ValueTypes.SET, ValueTypes.LIST] and field.datatype != Datatypes.JSON:
                return self.convert_value_from_db(field, value)

            elif field.valuetype == ValueTypes.MAP or field.valuetype == ValueTypes.LIST:
                return json.loads(value)

    def load_all(self):
        entities = []
        for unit_meta in self.__persisted_types:
            entities.extend(self.load_type(unit_meta))

        return entities

    def load_units(self):
        entities = []
        for unit_meta in EntityTypes:
            entities.extend(self.load_type(unit_meta))

        return entities

    def load_type(self, unit_meta, where_filters=None, load_referenced=False, entity_filter_method=None):
        entities = []

        if unit_meta.type_name[0] != "_":
            entities = self.load_entities(unit_meta, where_filters=where_filters, conn=None, entity_filter_method=entity_filter_method)

        if load_referenced:
            loaded_global_ids = set()
            requested_global_ids = set()
            loaded_global_ids.update([e.global_id() for e in entities])

            while True:
                walked_entities = Store.access(Operation.WALK, None, list(entities))
                load_map = {}

                for entity in walked_entities:
                    if entity.get_meta_type_enum() in self.__persisted_types and entity.committed == Committed.CREATED:
                        if not entity.global_id() in loaded_global_ids and not entity.global_id() in requested_global_ids:
                            load_map.setdefault(entity.get_meta_type_enum(), []).append(entity.entity_id)
                            requested_global_ids.add(entity.global_id())

                if not load_map:
                    break

                for meta_type, global_ids in load_map.items():
                    # Sub entities are always loaded, so do not apply entity_filter_method
                    loaded = self.load_entities(meta_type, where_filters=[("_entity_id", "in", global_ids)])
                    loaded_global_ids.update([e.global_id() for e in loaded])

        return entities

    def load_entity_ids_for_type(self, meta_type, where_filters=None, conn=None):
        entity_ids = []

        close_connection = ( conn is None )
        if conn is None:
            conn = self._create_connection()

        try:
            sql_tables = self.__sql_tables(self.__table_name(meta_type), meta_type.fields_enum)

            try:
                select_sql = self.__render_tempate(self.select_template, where_filters, sql_table=sql_tables[0], where_filters=where_filters)

                c = self._create_cursor(conn)

                print("{} with {}".format(select_sql, where_filters))
                if where_filters:
                    values = self.create_exectute_params()
                    for f in where_filters:
                        self.add_execute_param(values, f[0], f[2], f[1])
                    c.execute(select_sql, values)
                else:
                    c.execute(select_sql)

                rows = c.fetchall()
                for row in rows:
                    entity_ids.append(row[3])
            finally:
                self._close_cursor(c)

        finally:
            if close_connection:
                self._close_connection(conn)

        return entity_ids

    def load_entities(self, meta_type, where_filters=None, conn=None, entity_filter_method=None):
        entities = {}

        close_connection = ( conn is None )
        if conn is None:
            conn = self._create_connection()

        try:
            sql_tables = self.__sql_tables(self.__table_name(meta_type), meta_type.fields_enum)

            try:
                select_sql = self.__render_tempate(self.select_template, where_filters, sql_table=sql_tables[0], where_filters=where_filters)

                c = self._create_cursor(conn)

                #print("{} with {}".format(select_sql, where_filters))
                if where_filters:
                    values = self.create_exectute_params()
                    for f in where_filters:
                        self.add_execute_param(values, f[0], f[2], f[1])
                    c.execute(select_sql, values)
                else:
                    c.execute(select_sql)

                global_ids = set()

                rows = c.fetchall()
                for row in rows:
                    #print(row[0])
                    global_ids.add(row[0])
                    unit_global_id = row[1]
                    meta_type_global_id = row[2]
                    entity_id = row[3]
                    unit_entity = Store.access(Operation.GET, None, row[1])
                    meta_type_entity = Store.access(Operation.GET, None, row[2])
                    #print("{}:{}:{}".format(unit_entity, row[3], meta_type_entity))
                    entity = Store.access(Operation.GET, None, unit_entity, row[3], meta_type_entity)

                    index = 4 # 0 - global_id 1-3: base fields
                    for field in meta_type.fields_enum:
                        if not field.valuetype in [ValueTypes.LIST, ValueTypes.SET] or field.valuetype == ValueTypes.LIST and field.datatype == Datatypes.JSON:
                            value = self.map_value_from_db(field, row[index])
                            if not value is None:
                                entity.access(Operation.SET, value, field)

                            index += 1

                    entity.set_committed()
                    if entity_filter_method is None or entity_filter_method(entity):
                        entities[row[0]] = entity
                    elif entity_filter_method is not None:
                        entity.delete()
                        global_ids.remove(row[0])
            finally:
                self._close_cursor(c)

            # Do multivalue fields
            for field in meta_type.fields_enum:
                if field.valuetype in [ValueTypes.LIST, ValueTypes.SET] and field.datatype != Datatypes.JSON:
                    multivalue_sql_table = None
                    for stbl in sql_tables:
                        if stbl.table_name == "{}_{}".format(sql_tables[0].table_name, field.name):
                            multivalue_sql_table = stbl
                            break
                    multivalue_select_sql = self.__render_tempate(self.select_template, None, sql_table=multivalue_sql_table, filtwhere_filtersers=None)
                    try:
                        c = self._create_cursor(conn)
                        #print("{}".format(multivalue_select_sql))
                        c.execute(multivalue_select_sql)
                        rows = c.fetchall()
                        for row in rows:
                            if row[0] in global_ids:
                                entities[row[0]].access(Operation.ADD, self.map_value_from_db(field, row[2]), field)
                                entities[row[0]].set_committed()

                    finally:
                        self._close_cursor(c)
        finally:
            if close_connection:
                self._close_connection(conn)

        return entities.values()
    
    def __delete_updating_meta_records(self, delete_map):
        conn = None
        try:
            conn = self._create_connection()
            conn.autocommit = False
            committed_entities = []
            return_value = self.__update_entity_values(conn, delete_map, "delete", committed_entities)
            delete_map.clear()
            conn.commit()
            print("Committing meta deletion")
        except:
            traceback.print_exc()
            conn.rollback()
            return_value = False
        finally:
            self._close_connection(conn)

        return return_value

    def __prepare_delete_updating_meta_records(self, delete_map, table_name, where_filters):
        delete_sql = self.__render_tempate(self.delete_template, where_filters, sql_table=table_name, where_filters=where_filters)
        #print(delete_sql)
        delete_values = self.create_exectute_params(named_param=True)
        for where_filter in where_filters:
            self.add_execute_param(delete_values, name=where_filter[0], value=where_filter[2], operator=where_filter[1], named_param=True)
        delete_map[delete_sql] = [(None, delete_values)]

    def __prepare_update_entity_multivalues(self, entity, sql_tables, multivalues, update_map, delete_map=None):
        if len(sql_tables) > 1:
            for idx, sql_table in enumerate(sql_tables[1:], start=1):
                if not delete_map is None:
                    delete_sql = self.__render_tempate(self.delete_template, sql_table=sql_table)
                    delete_values = self.create_exectute_params(named_param=False)
                    self.add_execute_param(delete_values, "_global_id", entity.global_id(), named_param=False)
                    delete_map.setdefault(delete_sql, []).append((None, tuple(delete_values)))

                field_name, multivalues_array = multivalues[sql_table.table_name]
                if multivalues_array:
                    insert_sql = self.__render_tempate(self.insert_into_table_template, sql_table=sql_table)
                    for value_cnt, value in enumerate(multivalues_array):
                        values = self.create_exectute_params(named_param=False)
                        self.add_execute_param(values, "_global_id", entity.global_id(), named_param=False)
                        self.add_execute_param(values, "_value_cnt", value_cnt, named_param=False)
                        self.add_execute_param(values, "_"+field_name+"_value", value, named_param=False)
                        update_map.setdefault(insert_sql, []).append((None, tuple(values)))

    def __prepare_insert_entity(self, entity, insert_map):
        meta_type = entity.get_meta_type_enum()
        if meta_type in self.__persisted_types:
            sql_tables = self.__sql_tables(self.__table_name(meta_type), meta_type.fields_enum)
            multivalues = {}

            insert_sql = self.__render_tempate(self.insert_into_table_template, sql_table=sql_tables[0])
            values = self.create_exectute_params(named_param=False)
            self.add_execute_param(values, "_global_id", entity.global_id(), named_param=False)
            self.add_execute_param(values, "_unit", entity.unit.global_id(), named_param=False)
            self.add_execute_param(values, "_meta_type", entity.meta_type.global_id(), named_param=False)
            self.add_execute_param(values, "_entity_id", entity.entity_id, named_param=False)
            for field in meta_type.fields_enum:
                if not field.valuetype in [ValueTypes.LIST, ValueTypes.SET] or field.datatype == Datatypes.JSON and field.valuetype == ValueTypes.LIST:
                    self.add_execute_param(values, "_"+field.name, self.map_value_to_db(field, entity), named_param=False)
                else:
                    multivalue_tablename = "{}_{}".format(sql_tables[0].table_name, field.name)
                    multivalues[multivalue_tablename] = (field.name, self.map_value_to_db(field, entity))

            insert_map.setdefault(insert_sql, []).append((entity, tuple(values)))
            self.__prepare_update_entity_multivalues(entity, sql_tables, multivalues, insert_map)


    def __prepare_update_entity(self, entity, update_map, delete_map):
        meta_type = entity.get_meta_type_enum()
        if meta_type in self.__persisted_types:
            sql_tables = self.__sql_tables(self.__table_name(meta_type), meta_type.fields_enum)

            update_sql = self.__render_tempate(self.update_template, sql_table=sql_tables[0])
            values = self.create_exectute_params(named_param=False)

            multivalues = {}

            for field in meta_type.fields_enum:
                if not field.valuetype in [ValueTypes.LIST, ValueTypes.SET] or field.datatype == Datatypes.JSON and field.valuetype == ValueTypes.LIST:
                    self.add_execute_param(values, "_"+field.name, self.map_value_to_db(field, entity), named_param=False)
                else:
                    multivalue_tablename = "{}_{}".format(sql_tables[0].table_name, field.name)
                    multivalues[multivalue_tablename] = (field.name, self.map_value_to_db(field, entity))

            self.add_execute_param(values, "_global_id", entity.global_id(), named_param=False)

            update_map.setdefault(update_sql, []).append((entity, tuple(values)))
            self.__prepare_update_entity_multivalues(entity, sql_tables, multivalues, update_map, delete_map)

    def __is_entity_protected(self, e):
        meta_type = e.get_meta_type_enum()

        protected_unit_names = set()
        for type_initiated in _types_initiated:
            for type_initiated_meta_type in type_initiated:
                protected_unit_names.add(type_initiated_meta_type.unit_name)

        if meta_type in [EntityTypes.meta_field, EntityTypes.type_meta]:
            if e.unit.access(Operation.GET, None, UnitMeta.name) in protected_unit_names:
                #print(f"Protected {meta_type} in unit {e.unit.access(Operation.GET, None, UnitMeta.name)}. Not updating!")
                return True

        return False

    def persist_entities(self, entities, trx_state_callback=None):
        conn = None
        cnt = 0
        return_value = True
        committed_entities = []
        try:
            print("Persist {} entities".format(len(entities)))
            conn = self._create_connection()
            conn.autocommit = False

            thread_id = None
            if trx_state_callback:
                c = None
                try:
                    c = self._create_cursor(conn, prepared=False, buffered=False)
                    c.execute("SELECT CONNECTION_ID();")
                    thread_id = c.fetchone()[0]
                    print(f"CONNECTION ID is {thread_id}")
                    trx_state_callback(thread_id, TrxState.execute)
                finally:
                    if c is not None:
                        c.close()

            update_map = {}
            delete_map = {}
            cnt = 0
            for e in entities:
                if self.__is_entity_protected(e):
                    continue
                if e.committed == Committed.UPDATED:
                    self.__prepare_update_entity(e, update_map, delete_map)
                    cnt += 1
                
                if cnt % DB_UPDATE_BATCH == 0:
                    #print("Update: {}".format(cnt))
                    if return_value and delete_map:
                        return_value = self.__update_entity_values(conn, delete_map, "delete", committed_entities)
                        delete_map.clear()
                    if return_value and update_map:
                        return_value = self.__update_entity_values(conn, update_map, "update", committed_entities)
                        update_map.clear()
                    if not return_value:
                        raise Exception("Update failed")
                    cnt = 0

            if return_value and ( update_map or delete_map ):
                #print("Update: {}".format(cnt))
                if return_value and delete_map:
                    return_value = self.__update_entity_values(conn, delete_map, "delete", committed_entities)
                    delete_map.clear()
                if return_value and update_map:
                    return_value = self.__update_entity_values(conn, update_map, "update", committed_entities)
                    update_map.clear()
                if not return_value:
                    raise Exception("Update failed")

            insert_map = {}
            cnt = 0
            for e in entities:
                if self.__is_entity_protected(e):
                    continue
                if e.committed == Committed.CREATED:
                    self.__prepare_insert_entity(e, insert_map)
                    cnt += 1

                if return_value and insert_map and cnt % DB_UPDATE_BATCH == 0:
                    print("Insert: {}".format(cnt))
                    return_value = self.__update_entity_values(conn, insert_map, "insert", committed_entities)
                    insert_map.clear()
                    if not return_value:
                        raise Exception("Insert failed")
                    cnt = 0

            if return_value and insert_map:
                cnt = len(insert_map)
                print("Insert: {}".format(cnt))
                return_value = self.__update_entity_values(conn, insert_map, "insert", committed_entities)
                insert_map.clear()
                if not return_value:
                    raise Exception("Insert failed")


            if return_value:
                if thread_id:
                    trx_state_callback(thread_id, TrxState.commitstarted)
                conn.commit()
                if thread_id:
                    trx_state_callback(thread_id, TrxState.commitfinished)
                print("Committing {} entities".format(len(committed_entities)))
                for e in committed_entities:
                    e.set_committed()
            else:
                if thread_id:
                    trx_state_callback(thread_id, TrxState.rollbackstarted)
                conn.rollback()
                if thread_id:
                    trx_state_callback(thread_id, TrxState.rollbackfinished)

        except:
            traceback.print_exc()
            if thread_id:
                trx_state_callback(thread_id, TrxState.rollbackstarted)
            conn.rollback()
            if thread_id:
                trx_state_callback(thread_id, TrxState.rollbackfinished)
            return_value = False
        finally:
            self._close_connection(conn)

        return return_value

    def __update_entity_values(self, conn, map, update_type, committed_entities):
        print("Start executing {}: {}".format(update_type, len(map.keys())))
        sorted_keys = sorted(map, key=lambda k: len(map[k]))
        for sql in sorted_keys:
            try:
                value_array = map[sql]
                c = self._create_cursor(conn, prepared=True, buffered=False)
                start = time.time()
                print("SQL: {}, number of values: {}".format(sql, len(value_array)))
                for entity, values in value_array:
                    #if " entity_" in sql:
                    #    print(str(values))
                    c.execute(sql, values)
                    if entity:
                        committed_entities.append(entity)
                    if DELAY_IN_SEC_AFTER_UPDATE > 0.0001:
                        time.sleep(DELAY_IN_SEC_AFTER_UPDATE)                    
                end = time.time()
                print("Finished executing in {}".format(end-start))

            except:
                raise Exception("Update failed for sql {} - {}".format(sql, map[sql]))
            finally:
                self._close_cursor(c)

        return True

    def __render_tempate(self, template_func, *args, **kwargs):
        try: 
            template = Template(template_func(*args))
            return template.render(**kwargs)
        except:
            traceback.print_exc()

    def __table_name(self, unit_meta):
        return "{}_{}".format(self.__unit_meta_unit[unit_meta], unit_meta.type_name)

    def __sql_tables(self, table_name, fields_enum):
        sql_tables = []
        sql_table = SqlTable(table_name)
        sql_tables.append(sql_table)
        sql_table.add_field("_global_id", self.sql_type(Datatypes.STRING, ValueTypes.SINGLE, primary_key=True), primary_key=True, base_field=True)
        for base_field in EntityTypes._entity_base.fields_enum:
            if base_field != EntityBaseMeta.committed:
                sql_table.add_field("_"+base_field.name, self.sql_type(base_field.datatype, base_field.valuetype), base_field=True)
        for field in fields_enum:
            if field.valuetype in [ValueTypes.LIST, ValueTypes.SET] and field.datatype != Datatypes.JSON:
                multivalue_sql_table = SqlTable("{}_{}".format(table_name, field.name))
                multivalue_sql_table.add_field("_global_id", self.sql_type(Datatypes.STRING, ValueTypes.SINGLE, primary_key=True), primary_key=True, base_field=True)
                multivalue_sql_table.add_field("_value_cnt", self.sql_type(Datatypes.INT, ValueTypes.SINGLE), primary_key=True, base_field=True)
                multivalue_sql_table.add_field("_"+field.name+"_value", self.sql_type(field.datatype, ValueTypes.SINGLE))
                multivalue_sql_table.multi_value_table = True
                sql_tables.append(multivalue_sql_table)
                sql_table.value_tables_by_fields[field] = multivalue_sql_table
            else:
                sql_table.add_field("_"+field.name, self.sql_type(field.datatype, field.valuetype))

        return sql_tables

    def table_schemas(self, unit_meta, conn=None):

        table_schemas = []
        table_name = self.__table_name(unit_meta)

        created_tables = set()

        sql_tables = self.__sql_tables(table_name, unit_meta.fields_enum)
        for sql_table in sql_tables:
            if not self.__table_exists_internal(sql_table.table_name, conn):
                table_schemas.append(self.__render_tempate(self.create_table_template, sql_table, sql_table=sql_table))
                created_tables.add(sql_table)

        self.get_table_scema_changes(unit_meta, sql_tables, conn)

        for sql_table in sql_tables:
            if sql_table in created_tables:
                continue
            if sql_table.table_name in ALL_TABLES and ALL_TABLES[sql_table.table_name]["meta_sync_status"] == "VISITED":
                if ALL_TABLES[sql_table.table_name]["table_alterations"]:
                    # Add table alterations, so they can be executed
                    for table_alteration in ALL_TABLES[sql_table.table_name]["table_alterations"]:
                        match table_alteration["operation"]:
                            case "DELETE_MULTIVALUE_TABLE" | "DELETE_TABLE":
                                table_schemas.append(self.__render_tempate(self.delete_table_template, sql_table=sql_table))
                            case "CREATE_MULTIVALUE_TABLE":
                                table_schemas.append(self.__render_tempate(self.create_table_template, sql_table, sql_table=sql_table))

                if ALL_TABLES[sql_table.table_name]["field_alterations"]:
                    # Add table alterations, so they can be executed
                    field_modifications = []
                    for field_alterations in ALL_TABLES[sql_table.table_name]["field_alterations"]:
                        if "field_enum" in field_alterations:
                            field_enum = field_alterations["field_enum"]
                        match field_alterations["operation"]:
                            case "CHANGE_DATATYPE":
                                field_modifications.append(f"MODIFY {field_alterations['field_name']} {self.sql_type(Datatypes[field_enum['datatype']], ValueTypes[field_enum['valuetype']], False)}")
                            case "DELETE":
                                field_modifications.append(f"DROP {field_alterations['field_name']}")
                            case "ADD":
                                field_modifications.append(f"ADD {field_alterations['field_name']} {self.sql_type(Datatypes[field_enum['datatype']], ValueTypes[field_enum['valuetype']], False)}")

                    table_schemas.append(self.__render_tempate(self.alter_table_template, sql_table=sql_table, field_modifications=field_modifications))

                ALL_TABLES[sql_table.table_name]["meta_sync_status"] = "SQL_GENERATED"

        return table_schemas
    
    def __compare_table_fields_with_metadata(self, cursor, sql_table, multivalue_table, unit_meta, field_enum=None):
        table_exists = True
        table_specs = None
        try:
            if not ALL_TABLES:
                cursor.execute("SHOW TABLES")
                tables = []
                for table in cursor.fetchall():
                    ALL_TABLES.setdefault(table[0], {"table_name": table[0], "fields": [], "table_alterations": [], "field_alterations": [], "meta_sync_status": None})

                for table, table_specs in ALL_TABLES.items():
                    _multivalue_table = True
                    cursor.execute("DESCRIBE `" + table + "`;")
                    for field in cursor.fetchall():
                        if field[0] != "_global_id" and field[0] in BASE_FIELD_NAMES:
                            _multivalue_table = False
                        table_specs["fields"].append(field)

                    table_specs["multivalue_table"] = _multivalue_table

            table_specs = ALL_TABLES[sql_table.table_name]

        except:
            # Table does not exist
            table_exists = False
            table_specs = ALL_TABLES.setdefault(sql_table.table_name, {"table_name": sql_table.table_name, "fields": [], "table_alterations": [], "field_alterations": [], "meta_sync_status": None})

        if table_exists:
            existing_fields = []
            for field in table_specs["fields"]:
                if field[0] not in BASE_FIELD_NAMES:
                    existing_fields.append(field[0])
                    global_field_name = f"{unit_meta.unit_name}:{unit_meta.name}:{field[0][1:]}"
                
                    primary_key = field[3] == "YES"
                    if multivalue_table:
                        if field[0] not in MULTIVALUE_BASE_FIELD_NAMES:
                            sql_type = self.sql_type(field_enum.datatype, field_enum.valuetype, primary_key)
                            if sql_type.lower() != field[1].lower():
                                table_specs["field_alterations"].append({"operation": "CHANGE_DATATYPE", "field_name": field[0], "global_field_name": global_field_name, "field_enum": field_enum.to_json()})
                    else:
                        try:
                            meta_field = unit_meta.fields_enum[field[0][1:]]
                            sql_type = self.sql_type(meta_field.datatype, meta_field.valuetype, primary_key)
                            # Only existing fields get here, so if data type is changing or ValueType.SINGLE -> [ValueTypes.LIST, ValueTypes.SET] is possible, except if it is a JSON field
                            if sql_type.lower() != field[1].lower() or (meta_field.valuetype in [ValueTypes.LIST, ValueTypes.SET] and meta_field.datatype != Datatypes.JSON) :
                                if meta_field.valuetype == ValueTypes.SINGLE:
                                    table_specs["field_alterations"].append({"operation": "CHANGE_DATATYPE", "field_name": field[0], "global_field_name": global_field_name, "field_enum": meta_field.to_json()})
                                else:
                                    table_specs["field_alterations"].append({"operation": "DELETE", "field_name": field[0], "global_field_name": global_field_name, "field_enum": meta_field.to_json()})
                        except KeyError:
                            # field does not exist, delete it
                            table_specs["field_alterations"].append({"operation": "DELETE", "field_name": field[0], "global_field_name": global_field_name})

            if not multivalue_table:
                for meta_field in unit_meta.fields_enum:
                    if not f"_{meta_field.name}" in existing_fields and meta_field.valuetype not in [ValueTypes.LIST, ValueTypes.SET]:
                        global_field_name = f"{unit_meta.unit_name}:{unit_meta.name}:{meta_field.name}"
                        sql_type = self.sql_type(meta_field.datatype, meta_field.valuetype, primary_key)
                        table_specs["field_alterations"].append({"operation": "ADD", "field_name": f"_{meta_field.name}", "global_field_name": global_field_name, "field_enum": meta_field.to_json()})
        elif field_enum is not None:
            global_field_name = f"{unit_meta.unit_name}:{unit_meta.name}:{field_enum.name}"
            sql_type = self.sql_type(field_enum.datatype, field_enum.valuetype, False)
            table_specs["table_alterations"].append({"operation": "CREATE_MULTIVALUE_TABLE", "field_name": f"_{field_enum.name}", "global_field_name": global_field_name, "field_enum": field_enum.to_json()})

        if ALL_TABLES[sql_table.table_name]["meta_sync_status"] is None:
            ALL_TABLES[sql_table.table_name]["meta_sync_status"] = "VISITED"

    def print_table_alterations(self, unit_name, only_alter=True):
       table_alerations = [{"table_name": f["table_name"], "table_alterations": f["table_alterations"], "meta_sync_status": f["meta_sync_status"]} for f in ALL_TABLES.values() if ( not only_alter or f["table_alterations"] ) and f["table_name"].startswith(f"{unit_name}_")]
       if table_alerations:
            print(json.dumps(table_alerations, indent=4))
    

    def get_table_scema_changes(self, unit_meta, sql_tables, conn=None):
        close_connection = conn is None
        try:
            if conn is None:
                conn = self._create_connection()
            c = self._create_cursor(conn, buffered=False)
            # Get fields:
            self.__compare_table_fields_with_metadata(c, sql_tables[0], False, unit_meta)

            for field, multivalue_sql_table in sql_tables[0].value_tables_by_fields.items():
                self.__compare_table_fields_with_metadata(c, multivalue_sql_table, True, unit_meta, field)

            self._close_cursor(c)

        except:
            traceback.print_exc()
        finally:
            if close_connection:
                self._close_connection(conn)

    def __table_exists_internal(self, table_name, conn=None):
        if conn is None:
            try:
                conn = self._create_connection()
                return self.table_exits(table_name, conn)
            finally:
                self._close_connection(conn)
        else:
            return self.table_exits(table_name, conn)

    def __generate_base_schema(self, conn=None):
        return self.generate_schema(UNIT_ENTITY, EntityTypes, conn)
        #self.generate_schema(UNIT_EVENTS, EventTypes, conn)

    def generate_schema(self, unit_name, unit_meta_enums, conn = None):
        close_connection = ( conn is None )
        schema = []

        if conn is None:
            conn = self._create_connection()

        try:
            for unit_meta in unit_meta_enums:
                self.__unit_meta_unit[unit_meta] = unit_name
                self.__persisted_types.add(unit_meta)
                if unit_meta.type_name[0] != "_":
                    table_name = self.__table_name(unit_meta)
                    tbl_schema_strings = self.table_schemas(unit_meta, conn)
                    for tbl_schema_string in tbl_schema_strings:
                        if not tbl_schema_string is None:
                            print(tbl_schema_string)
                            schema.append(tbl_schema_string)
                            self.alter_table(tbl_schema_string, conn)
            if not EntityTypes.unit in self.__persisted_types:
                schema.extend(self.__generate_base_schema(conn))

            for table_specs in ALL_TABLES.values():
                if table_specs["table_name"].startswith(f"{unit_name}_") and table_specs["meta_sync_status"] is None:
                    sql_table = SqlTable(table_specs["table_name"])
                    tbl_schema_string = self.__render_tempate(self.delete_table_template, sql_table=sql_table)
                    print(tbl_schema_string)
                    schema.append(tbl_schema_string)
                    self.alter_table(tbl_schema_string, conn)
                    #table_specs["table_alterations"].append({"table_name": table_specs["table_name"], "operation": "DELETE_MULTIVALUE_TABLE" if table_specs["multivalue_table"] else "DELETE_TABLE"})
                    #table_specs["meta_sync_status"] = "VISITED"
        finally:
            if close_connection:
                self._close_connection(conn)

        #for sch in schema:
        #    print(sch)

        return schema
    
    def dump_database(self, stream):
        conn = self._create_connection()

        try:

            c = self._create_cursor(conn, buffered=False)

            try:
                max_packet_size = int(os.environ.get("MYSQL_MAX_ALLOWED_PACKET_SIZE", DEFAULT_MAX_PACKET_SIZE))
            except:
                max_packet_size = DEFAULT_MAX_PACKET_SIZE

            mysql_version = ""

            c.execute("SHOW VARIABLES LIKE 'version'")
            for vars in c.fetchall():
                mysql_version = vars[1]

            stream.write("-- MySQL dump (pydust)\n"+
                         "--\n"+
                         f"-- Host: {os.environ.get('HOSTNAME')}    Database: {os.environ.get('MYSQL_DB')}\n"+
                         "-- ------------------------------------------------------\n"+
                         f"-- Server version	{mysql_version}\n\n")

            stream.write("/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT */;\n"+
                         "/*!40101 SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS */;\n"+
                         "/*!40101 SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION */;\n"+
                         "/*!50503 SET NAMES utf8mb4 */;\n"+
                         "/*!40103 SET @OLD_TIME_ZONE=@@TIME_ZONE */;\n"+
                         "/*!40103 SET TIME_ZONE='+00:00' */;\n"+
                         "/*!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 */;\n"+
                         "/*!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 */;\n"+
                         "/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO' */;\n"+
                         "/*!40111 SET @OLD_SQL_NOTES=@@SQL_NOTES, SQL_NOTES=0 */;\n\n")

            c.execute("SHOW TABLES")
            tables = []
            for table in c.fetchall():
                tables.append(table[0])

            # Lower the packet size slightly, just in case
            max_packet_size -= 1000

            for table in tables:
                stream.write("--\n"+
                         f"-- Table structure for table `{table}`\n"+
                         "--\n\n")
                stream.write("DROP TABLE IF EXISTS `" + str(table) + "`;\n")

                stream.write("/*!40101 SET @saved_cs_client     = @@character_set_client */;\n"+
                         "/*!50503 SET character_set_client = utf8mb4 */;")

                c.execute("SHOW CREATE TABLE `" + str(table) + "`;")
                stream.write("\n" + str(c.fetchone()[1]) + ";\n");

                stream.write("/*!40101 SET character_set_client = @saved_cs_client */;\n\n")

                stream.write("--\n"+
                         f"-- Dumping data for table `{table}`\n"+
                         "--\n\n"+
                         f"LOCK TABLES `{table}` WRITE;\n"+
                         f"/*!40000 ALTER TABLE `{table}` DISABLE KEYS */;\n")

                # Get fields:
                c.execute("DESCRIBE `" + str(table) + "`;")
                fields = []
                for field in c.fetchall():
                    field_type = field[1]
                    if field_type.lower() in ["binary", "varbinary", "blob", "mediumblob", "longblob"]:
                        fields.append((f"HEX({field[0]})", field[0], True)) 
                    else:
                        fields.append((field[0], field[0], False)) 


                c.execute("SELECT {} FROM `{}`;".format(",".join(f[0] for f in fields), table))
                row = c.fetchone()

                row_index = 0
                packet_length  = 0

                empty_table = True

                if row is not None:
                    empty_table = False
                    insert_str = "INSERT INTO `{}` ({}) VALUES ".format(table, ",".join(f[1] for f in fields))
                    packet_length = len(insert_str)
                    stream.write(insert_str)
                    first_row = True

                while row is not None:
                    row_stream = io.StringIO("")

                    if not first_row:
                        row_stream.write(",")

                    row_stream.write("(")
                    first = True
                    for field_idx in range(len(fields)):
                        field = fields[field_idx]
                        if not first:
                            row_stream.write(",");
                        if row[field_idx] is None:
                            row_stream.write("NULL")
                        elif field[2]:
                            row_stream.write(f"UNHEX(\"{row[field_idx]}\")")
                        elif isinstance(row[field_idx], str):
                            escaped_value = row[field_idx].replace('\\','\\\\').replace('"','\\"')
                            row_stream.write(f"\"{escaped_value}\"")
                        else:
                            row_stream.write(f"\"{row[field_idx]}\"")
                        first = False
                    row_stream.write(")")

                    row_str = row_stream.getvalue()
                    if packet_length + len(row_str) > max_packet_size:
                        stream.write(";\n")
                        insert_str = "INSERT INTO `{}` ({}) VALUES ".format(table, ",".join(f[1] for f in fields))
                        packet_length = len(insert_str)
                        stream.write(insert_str)
                        if not first_row:
                            row_str = row_str[1:]
                    
                    stream.write(row_str)
                    packet_length += len(row_str)
                    first_row = False

                    row = c.fetchone()
                    row_index += 1

                if not empty_table:
                    stream.write(";\n")

                stream.write(f"/*!40000 ALTER TABLE `{table}` ENABLE KEYS */;\n"+
                         "UNLOCK TABLES;\n\n")

            stream.write("/*!40101 SET SQL_MODE=@OLD_SQL_MODE */;\n"+
                         "/*!40014 SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS */;\n"+
                         "/*!40014 SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS */;\n"+
                         "/*!40101 SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT */;\n"+
                         "/*!40101 SET CHARACTER_SET_RESULTS=@OLD_CHARACTER_SET_RESULTS */;\n"+
                         "/*!40101 SET COLLATION_CONNECTION=@OLD_COLLATION_CONNECTION */;\n"+
                         "/*!40111 SET SQL_NOTES=@OLD_SQL_NOTES */;\n\n"+
                         f"-- Dump completed on {datetime.datetime.strftime(datetime.datetime.now(), FORMAT_DB_DATETIME)}\n")

            self._close_cursor(c)
        except:
            traceback.print_exc()
        finally:
            self._close_connection(conn)
